import threading
import sys, os
import ctypes
from collections import OrderedDict
import logging
import importlib
import pprint
from pathlib import Path

import numpy as np

from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtCore import Qt

from .. import config, gui, __release__
from ..core import conf

from .base import BasePanel
from .window import MainWindow
from ..ezdock.ezdock import DockManager
from ..ezdock.laystruct import LayoutStruct
from ..utils import new_id_using_keys
from ..gcore.utils import getMenuAction

respath = Path(config['respath'])
sck = config['shortcuts']
logger = logging.getLogger(__name__)

class Panels(object):
    def __init__(self, qapp):
        self.panels = OrderedDict()
        self.ezm = DockManager(self, qapp)
        self.qapp = qapp

    def keys(self):
        return self.panels.keys()

    def __getitem__(self, category):
        return self.panels[category]

    def id_exists(self, category, panid):
        if not category in self.keys():
            self.panels[category] = OrderedDict()

        return panid in self.panels[category].keys()

    def new_id(self, category):
        return int(new_id_using_keys(tuple(self[category].keys())))

    def items(self):
        for item in self.panels.items():
            yield item

    def move_to_end(self, widget, category=None):
        if not category is None:
            catpanels = self.panels[category]
            panid = next((k for k,v in panels.items() if v is widget), None)

        else:
            for category, catpanels in self.panels.items():
                panid = next((k for k,v in catpanels.items() if v is widget), None)
                if not panid is None: break
            else:
                panid = None
                category = None

        if not panid is None:
            catpanels.move_to_end(panid)
            self.panels.move_to_end(category)

        return panid, category

    def __iter__(self):
        for category in self.keys():
            yield category, self[category]

    def get_active_panid(self, category, panidpos=-1):
        panel = self.selected(category, panidpos)
        if panel is None:
            return None
        else:
            return panel.panid


    def new(self, category, paneltype=None, windowname=None, *args, **kwargs):
        image_classes = self.classes_of_category(category)
        if paneltype is None:
            ImageClass = next(iter(image_classes.values()))
        else:
            ImageClass = image_classes[paneltype]
        panel = self.new_panel(ImageClass, windowname, None, args=args, kwargs=kwargs)
        return panel


    def select_or_new(self, category, panid=None, defaulttype='basic', parentName='main', args=(), kwargs={}):
        """
        If panid < 0, -1: select the active panel, -2: selected before that, ...
        panid > 0: select the panel if exists, otherwise a new with that number
        """

        if not panid is None and panid < 0:
            panel = self.selected(category, panid)

        elif category in self.keys():
            panel = self[category].get(panid, None)

        else:
            panel = None

        if panid is not None and panid < 0:
            panid = None

        if panel is None:
            panel = self.new(category, defaulttype, parentName, *args, **kwargs)

        panel.select()

        return panel


    def selected_category(self):
        return tuple(self.panels.keys())[-1]


    def selected(self, category, panidpos=-1, panel=True):
        assert panidpos < 0

        if category in self.keys():
            panids = tuple(self[category].keys())
        else:
            return None

        if abs(panidpos) <= len(panids):
            panid = panids[panidpos]
            if panel:
                panel = self[category][panid]
                return panel
            else:
                return panid
        else:
            return None


    def reselect_all(self):
        for category in list(self.keys()):
            panel = self.selected(category)
            if not panel is None:
                #logger.info(f'Selecting {category}: {panel.panid}')
                panel.select()

    def restore_state_from_config(self, layout_name):
        if isinstance(layout_name, int):
            layout_name = config['shortcuts']['layout'][str(layout_name)]
        perspective = config['layout'][layout_name]
        self.ezm.set_perspective(perspective)

    def classes_of_category(self, category):
        panelClasses = BasePanel.userPanelClasses()
        panelClassesCat = panelClasses.get(category, [])
        return dict([(Cls.panelShortName, Cls) for Cls in panelClassesCat])

    def new_panel(self, PanelClass, parentName=None, panid=None, floating=False, position=None, args=(), kwargs={}):

        if parentName is None:
            activeWindow = self.qapp.activeWindow()
            if isinstance(activeWindow, MainWindow):
                parentName =  activeWindow.name

        panel = PanelClass(parentName, panid, *args, **kwargs)
        panel.show()

        if floating:
            window = panel
        else:
            window = self.ezm.new_window_on_panel(panel, parentName)

        if position is None:
            position = self.place_window(window, panel.category)

        window.move(position)
        panel.select()

        return panel

    def place_window(self, window, category):
        desktop_rect = QtWidgets.QDesktopWidget().availableGeometry()
        window_rect = window.frameGeometry()
        prior_panel = self.selected(category, -2)

        if not prior_panel is None:
            prior_rect = prior_panel.frameGeometry()
            topleft = prior_panel.mapToGlobal(QtCore.QPoint(0,0))
            prior_rect.moveTopLeft(QtCore.QPoint(topleft.x()+10, topleft.y()+10))
            center = prior_rect.center()
        else:
            center = desktop_rect.center()

        window_rect.moveCenter(center)
        position = window_rect.topLeft()

        if not desktop_rect.contains(window_rect):
            position = QtCore.QPoint(0,0)

        return position

    def get_menu_action(self, category, panid, menutrace):
        """
        Trigger a menu action of a panel.

        :param str category: Example 'image'
        :param int id: Example 1
        :param list menutrace: Example ['File', 'New Image']
        """

        if category is None:
            window = self.qapp.activeWindow()
            if not isinstance(window, MainWindow):
                raise KeyError('Action not found')
            category = window.activeCategory

        if panid is None:
            panid = self.get_active_panid(category)
            
        if panid is None:
            return

        panel = self[category][panid]

        return getMenuAction(panel.menuBar(), menutrace)


    def removeBindingsTo(self, category, panid):
        for c, panels in self.items():
            for p, panel in panels.items():
                if (category, panid) in panel.bindings:
                    panel.removeBindingTo(category, panid)



"""multiple dimension range and slicing

    virtual sequence of numbers defined by start to stop by step
    functions to
        * modify the range by applying slices
        * reset to full range
        * get default python range object
        * partial ranges for representations

    start = first number; 0 by default
    stop = all numbers are lower then stop, so not included in sequence
    step = stepping of the sequence; 1 by default
    maxstop = stop can never be set larger than maxstop

    DimRange: single dimension
    DimRanges: multiple dimensions
"""


class DimRange(object):
    """range info of a single dimension

    maxstop = the size of the dimension

    apply slices to a range to get a new range
    slices can be applied cumulatively
    range can be reset to the original range
    """
    def __init__(self, maxstop):
        self.maxstop = maxstop
        self.start = 0
        self.stop = maxstop
        self.step = 1

    def reset(self):
        self.start = 0
        self.stop = self.maxstop
        self.step = 1

    def clone(self):
        return DimRange(self.maxstop)

    def inherite(self, source):
        self.start = source.start
        self.stop = source.stop
        self.step = source.step

    def copy(self):
        result = self.clone()
        result.inherite(self)
        return result

    def setfromslice(self, aslice, full=True):
        """slice the current range by the slice aslice

        a slice is e.g. created from item[0:10:2]
            or by slice(0,10,2)

        if full == True: don't build on the current roi
        """
        if full:
            # reset the roi before the new roi
            self.reset()

        prior_start = self.start
        prior_stop = self.stop
        prior_step = self.step

        if isinstance(aslice, int):
            # the slice is just one index
            if aslice < 0:
                aslice = max(prior_stop + prior_step * aslice, 0)
            aslice = slice(aslice, aslice+1, None)

        if aslice.start is None:
                self.start = prior_start
        elif aslice.start < 0:
            self.start = max(prior_stop + prior_step * aslice.start, 0)
        else:
            self.start = min(prior_start + prior_step * aslice.start, prior_stop)

        if aslice.stop is None:
            self.stop = prior_stop
        elif aslice.stop < 0:
            self.stop = max(prior_stop + prior_step * aslice.stop, 0)
        else:
            self.stop = min(prior_start + prior_step * aslice.stop, prior_stop)

        if aslice.step is None:
            self.step = prior_step
        else:
            self.step = prior_step * aslice.step

    def applyslice(self, aslice, full=False):
        """return a range based on another range and a range definition

        ex:
           bslice = slice(3,None,2) or [3::2] on the __getitem__ or __setitem__
           brange = (0,50,2)
           return range(6,50,4)
        """
        self.setfromslice(aslice, full)

    def clip(self):
        if self.start < self.maxstop:
            self.start = max(self.start, 0)
        else:
            self.start = 0
        if self.stop > 0:
            self.stop = min(self.stop, self.maxstop)
        else:
            self.stop = self.maxstop
            
    def size(self):
        return len(self)

    def __len__(self):
        return len(self.range)

    def isfullrange(self):
        if (self.start == 0) and (self.stop == self.maxstop) and (self.step == 1):
            return True
        else:
            return False

    def getstartrange(self, count=7):
        half = count // 2
        if self.count > count:
            tmp = self.copy()
            tmp.applyslice(slice(None, half, None), False)  # 2nd argument unnecessary
            return tmp.getrange()
        else:
            return self.getrange()

    def getstoprange(self, count=7):
        half = count // 2
        if self.count > count:
            tmp = self.copy()
            if (tmp.stop - tmp.start) % tmp.step != 0:
                tmp.maxstop += tmp.step - (tmp.stop - tmp.start) % tmp.step
                tmp.stop = tmp.maxstop
            tmp.applyslice(slice(-half, None,  None), False)  # 2nd argument unnecessary
            return tmp.getrange()
        else:
            return None

    def __str__(self):
        return 'dim:%d slice:%d:%d:%d' % (self.maxstop, self.start, self.stop, self.step) 

    def __repr__(self):
        return str(self)

    def __getitem__(self, index):
        newrange = self.copy()
        newrange.applyslice(index, False)  # 2nd argument is unnecessary
        return newrange

    def getslice(self):
        return slice(self.start, self.stop, self.step)        
        
    def getrange(self):
        return range(self.start, self.stop, self.step)
        
    def splitstarts(self, count):
        from math import ceil
        l = len(self)
        step = ceil(l / count)
        return self.range[::step]

    count = property(size)
    slice = property(getslice)
    range = property(getrange)


class SetSlices(object):
    """
    to set the roi, use the slices syntax on set
    example:
        b = pybix.ones(100,80) * 12000
        b.roi
        >>>[dim:100 slice:0:100:1, dim:80 slice:0:80:1]
        b.roi.set[20:60,::2]
        b.roi
        >>>[dim:100 slice:20:60:1, dim:80 slice:0:80:2]            
    """
    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, indices):
        self.parent.setslices(indices)
        
    def __repr__(self):     
        return self.__doc__


class DimRanges(object):
    """range info of a multiple dimensions

    sizes = tuple of the sizes of the dimensions
    """
    def __init__(self, sizes=None):
        self.rngs = []
        self.set = SetSlices(self)
        if sizes is not None:
            for size in sizes:
                self.rngs.append(DimRange(size))

    def maketuple(self, index):
        if not isinstance(index, tuple):
            return (index,)
        else:
            return index

    def setslices(self, index):
        slices = self.maketuple(index)
        for i in range(len(self.rngs)):
            self.rngs[i].applyslice(slices[i], full=True)

    def accslices(self, index):
        slices = self.maketuple(index)
        for (rng, slc) in zip(self.rngs, slices):
            rng.applyslice(slc, full=False)

    def getslices(self, swap_row_columns=False):
        tmp = []        
        for rng in self.rngs:
            tmp.append(rng.getslice())
        if swap_row_columns:
            # swap first and second index
            tmp[0], tmp[1] = tmp[1], tmp[0]
        return tuple(tmp)

    def getshape(self):
        tmp = []
        for rng in self.rngs:
            tmp.append(len(rng))
        return tuple(tmp)        
    
    shape = property(getshape)
    
    def getfullshape(self):
        tmp = []
        for rng in self.rngs:
            tmp.append(rng.maxstop)
        return tuple(tmp)

    fullshape = property(getfullshape)

    def reset(self):
        for rng in self.rngs:
            rng.reset()

    def clip(self):
        for rng in self.rngs:
            rng.clip()

    def clone(self):
        return DimRanges(self.shape)

    def inherite(self, source):
        for (tarr, srcr) in zip(self.rngs, source.rngs):
            tarr.inherite(srcr)

    def copy(self):
        newrngs = self.clone()
        newrngs.inherite(self)
        return newrngs

    def isfullrange(self):
        test = True
        for rngs in self.rngs:
            test &= rngs.isfullrange()
        return test        

    def __len__(self):
        length = 1
        for rng in self.rngs:
            length *= len(rng)
        return length

    def __getitem__(self, indices):
        newranges = self.copy()
        newranges.accslices(indices)
        return newranges

    def __str__(self):
        return str(self.rngs)

    def __repr__(self):
        return str(self)



import collections
import importlib
import pprint
import logging
import pathlib

from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy import QtCore, QtGui, QtWidgets

from .. import gui, config
from .laystruct import LayoutStruct
from .dockwidgets import DockContainer, DockTab, DockTag

respath = pathlib.Path(config['respath'])
logger = logging.getLogger(__name__)

class HoverButton(QPushButton):
    icons = dict()
        
    def __init__(self, caption, parent):                
        if caption in HoverButton.icons.keys():
            super().__init__('', parent)
            self.setIcon(HoverButton.icons[caption])
        else:
            super().__init__(caption, parent)
        
        self.setAcceptDrops(True)
        
    @staticmethod
    def load_icons():
        HoverButton.icons['T'] = QtGui.QIcon(str(respath / 'icons' / 'dock_tab.png'))
        HoverButton.icons['L'] = QtGui.QIcon(str(respath / 'icons' / 'dock_left.png'))
        HoverButton.icons['R'] = QtGui.QIcon(str(respath / 'icons' / 'dock_right.png'))
        HoverButton.icons['U'] = QtGui.QIcon(str(respath / 'icons' / 'dock_up.png'))
        HoverButton.icons['B'] = QtGui.QIcon(str(respath / 'icons' / 'dock_bottom.png'))

    def enterEvent(self, event):
        self.startPreview()
        
    def leaveEvent(self, event):
        self.endPreview()        

    def dragEnterEvent(self, event):
        event.accept()
        self.startPreview()      
        
    def dragLeaveEvent(self, event):
        self.endPreview()

    def startPreview(self):
        self.parent().active_rect_index = self.rect_id        
        self.parent().repaint()
        
    def endPreview(self):
        self.parent().endPreview()      
        
    def dropEvent(self, event):
        position = event.pos()
        event.setDropAction(Qt.MoveAction)
        event.accept()  
        self.clicked.emit()


class DockOverlay(QWidget):
    def __init__(self, parent, tool=False):            
        self.tool = tool
        
        if self.tool:
            super().__init__(parent=None)                       
            self.container = parent
            self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint)        
            self.setAttribute(Qt.WA_NoSystemBackground)
            self.setAttribute(Qt.WA_TranslucentBackground)
            self.setWindowOpacity(1)
        
        else:        
            super().__init__(parent=parent)
            self.container = parent
            palette = QPalette(self.palette())
            palette.setColor(palette.Background, Qt.transparent)
            self.setPalette(palette)        
        
        #self.setMouseTracking(True)                        
        #self.rects = []
        self.active_rect_index = None
     
        self.hide()
        
        #font = QtWidgets.QApplication.instance().font()
        fontmetric = QtGui.QFontMetrics(self.font())
        fontheight = fontmetric.height()
        
        btnsize = int(round(fontheight * 2.5)) // 2 * 2
        
        #Size parameters of the buttons
        self.btnw = btnsize 
        self.btnh = btnsize
        self.btnb = btnsize // 2 + 2
        self.btnc = btnsize + 2
        
    @property
    def ezm(self):
        return self.container.manager        
        
    def show_and_insert(self, window, insert_node):
        self.insert_window = window
        self.insert_node = insert_node  
        container = self.container
        self.match_geometry(container)
        self.get_dock_and_button_positions(container)
        self.show()
        
    def find_best_fit(self, lower_container, upper_container):
        self.match_geometry(lower_container)
        rects = self.get_dock_positions(lower_container)
        
        pos = self.mapFromGlobal(upper_container.mapToGlobal(upper_container.pos()))
        x, y = pos.x(), pos.y()
        w, h = upper_container.width(), upper_container.height()
        
        diffs = []
        for tags_rect in rects:
            tags, rect = tags_rect
            caption, panqualid, btnpos = tags            
            rx, ry, rw, rh = rect[0], rect[1], rect[2], rect[3]
            diff = sum(((rx - x)**2         , (ry - y)**2,
                       (rx + rw - x - w)**2, (ry - y)**2, 
                       (rx - x)**2         , (ry + rh - y - h)**2,
                       (rx + rw - x - w)**2, (ry + rh - y - h)**2))
            diffs.append((len(diffs), diff))
                
        diffs = sorted(diffs, key= lambda item: item[1])
        lowest = diffs[0][0]
        return rects[lowest]
        
    def get_dock_and_button_positions(self, container):
        self.rects = self.get_dock_positions(container)
    
        self.buttons = []
        
        w, h = self.btnw, self.btnh
        
        for index, tags_rect in enumerate(self.rects):
            tags, rect = tags_rect
            caption, panqualid, pos = tags
            button = HoverButton(caption, self)                                    
            button.setGeometry(pos[0]- w//2, pos[1]- h//2, w, h)                               
            button.clicked.connect(self.endOverlay)
            button.rect_id = index
            
            self.buttons.append(button)

    def get_dock_positions(self, container):        
        rects = []
        
        spaceb = self.btnb
        spacec = self.btnc
        
        geo = self.geometry() 
        w, h = geo.width(), geo.height()
        rects.append((('U', ('top', None),    (w/2, spaceb  )), (0, 0    , w,   h/3)))
        rects.append((('B', ('bottom', None), (w/2, h-spaceb)), (0, 2*h/3, w,   h/3)))
        rects.append((('L', ('left', None),   (spaceb,  h/2 )), (0, 0,     w/3, h  )))
        rects.append((('R', ('right', None),  (w-spaceb,h/2 )), (2*w/3, 0, w/3, h  )))   
        
        panelrects = self.get_panel_positions(container)    

        for panel, rect in panelrects:
            panqualid = (panel.category, panel.panid)            
            pos = self.mapFromGlobal(QtCore.QPoint(rect[0], rect[1]))
            posx = pos.x()
            posy = pos.y()
            w, h = rect[2], rect[3]
            xc, yc = posx + w/2, posy + h/2
            rects.append((('T', ('tab', panqualid),    (xc, yc   )), (posx, posy, w, h)))
            rects.append((('U', ('top', panqualid),    (xc, yc-spacec)), (posx, posy, w, h/2)))
            rects.append((('B', ('bottom', panqualid), (xc, yc+spacec)), (posx, posy + h/2, w, h/2)))
            rects.append((('L', ('left', panqualid),   (xc-spacec, yc)), (posx, posy, w/2, h)))
            rects.append((('R', ('right', panqualid),  (xc+spacec, yc)), (posx  + w/2, posy, w/2, h)))     

        return rects          

    def get_panel_positions(self, container):
        rects = []   
        
        for cat, panid in container.panelIds:
            panel = gui.qapp.panels[cat][panid]
            #TO DO: if panel is part of scrollarea, it is possible
            # that the panel is not completly visible
            # So take the visible part. But how?
            if not panel.isVisible():
                continue
            pos = panel.mapToGlobal(panel.pos())
            rect = (panel, (pos.x(), pos.y(), panel.width(), panel.height()))
            rects.append(rect)
            
        return rects
        
    def match_geometry(self, window):
        if self.tool:
            geo = window.geometry()                        
            pos = window.mapToGlobal(geo.topLeft())
            self.setGeometry(pos.x(), pos.y(), geo.width(), geo.height())
        else:
            geo = window.geometry()
            w, h = geo.width(), geo.height()
            self.setGeometry(0, 0, w, h)
        
    def endOverlay(self):
        if not self.active_rect_index is None:
            tags, rect = self.rects[self.active_rect_index]     
            caption, panqualid, pos = tags            
            
        self.ezm.hide_overlays()      
        self.place_node(*panqualid)  

    def endPreview(self):
        self.active_rect_index = None       
        self.repaint()         
        
    def mousePressEvent(self, event):
        self.ezm.hide_overlays()
        geo = self.insert_window.geometry()
        geo.moveTo(QCursor.pos())
        self.insert_window.setGeometry(geo)   
        self.insert_window.show()

    def paintEvent(self, event):
        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.Antialiasing)
        
        painter.fillRect(event.rect(), QBrush(QColor(255, 255, 255, 127)))
        
        if not self.active_rect_index is None:
            position, rect = self.rects[self.active_rect_index]
            painter.fillRect(*rect, QBrush(QColor(0, 255, 0, 127)))          
            
        painter.setPen(QPen(Qt.NoPen))
        
    def place_node(self, relative_pos, refpanqualid):  
        ls = self.container.get_layout_struct()
        ls.compact()
        ls.insert_branch(self.insert_node, relative_pos, refpanqualid)   
        #ls.compact()
        self.insert_window.container.update_layout(LayoutStruct())
        self.container.update_layout(ls)


import sys
import numpy as np

from ... import gui
    
def spectrogram(arr, vertical=False, plot=True):
    """
    Calculates the fullnoise, whitenoise and whiteness.
    and plot a spectrogram.
    
    https://www.emva.org/wp-content/uploads/EMVA1288-3.1a.pdf
    
    :param np.ndarray arr: A 2 dimensional array
    :param bool vertical: Calculate in vertical direction
    :param bool plot: Plot and print
    :returns: fullnoise, whitenoise, whiteness
    :rtype: tuple(float, float, float)
    """
    if vertical:
        return spectr_vert(arr, plot)
    else:
        return spectr_hori(arr, plot)
    
def spectr_hori(arr, plot=True):            
    """
    Calculates the fullnoise, whitenoise and whiteness.
    and plot a horizontal spectrogram.
    
    :param np.ndarray arr: A 2 dimensional array
    :returns: fullnoise, whitenoise, whiteness
    :rtype: tuple(float, float, float)    
    """
    arr = arr.astype('double')
    ydim, xdim = arr.shape
    arr -= arr.mean()
    mag = abs(np.fft.fft(arr,axis=1)) / xdim ** 0.5
    spr = (np.sum(mag**2,0) / ydim) ** 0.5
    fullnoise = (np.sum(spr**2) / (xdim+1))** 0.5
    whitenoise = np.median(spr)
    whiteness = (fullnoise / whitenoise)
    
    if plot:
        plt = gui.prepareplot()
        plt.figure()
        plt.grid(True)
        plt.title('Horizontal Spectrogram')
        plt.plot(spr)    
    
        print("Fullnoise    :  %8.2f" % fullnoise)
        print("WhiteNoise   :  %8.2f" % whitenoise)
        print("The whiteness:  %8.2f Ideal this is 1" % whiteness) 
    
    return fullnoise, whitenoise, whiteness
    
def spectr_vert(arr, plot=True): 
    """
    Calculates the fullnoise, whitenoise and whiteness.
    and plot a vertical spectrogram.
    
    :param np.ndarray arr: A 2 dimensional array
    :returns: fullnoise, whitenoise, whiteness
    :rtype: tuple(float, float, float)    
    """
    arr = arr.astype('double')
    ydim, xdim = arr.shape
    arr -= arr.mean()
    mag = abs(np.fft.fft(arr,axis=0)) / ydim ** 0.5
    spr = (np.sum(mag**2,1) / xdim) ** 0.5
    fullnoise = (np.sum(spr**2) / (ydim+1))** 0.5
    whitenoise = np.median(spr)
    whiteness = (fullnoise / whitenoise)
    
    if plot:
        plt = gui.prepareplot()
        plt.figure()
        plt.grid(True)
        plt.title('Vertical Spectrogram')
        plt.plot(spr)            
    
        print("Fullnoise    :  %8.2f" % fullnoise)
        print("WhiteNoise   :  %8.2f" % whitenoise)
        print("The whiteness:  %8.2f Ideal this is 1" % whiteness) 
    
    return fullnoise, whitenoise, whiteness    


import sys
import time
import logging
import pickle
from pathlib import Path

from matplotlib.backends.backend_qt5agg import  FigureCanvas, NavigationToolbar2QT as NavigationToolbar
from matplotlib._pylab_helpers import Gcf
from matplotlib.backends.backend_template import FigureCanvasTemplate, FigureManagerTemplate 
import pylab

from qtpy import QtWidgets, QtCore, QtGui              

from ... import gui, config
from .. import BasePanel, CheckMenu

respath = Path(config['respath'])
logger = logging.getLogger(__name__)

set_active_backup = Gcf.set_active

def gcf_set_active(manager):
    # Note that this function is called by pyplot
    # - when the user clicks on the canvas (callback needed on canvas object?)
    # - on pyplot.figure(number)
    if pylab.get_backend() == config.get("matplotlib", {}).get("backend", None):
        gui.plot.select(manager.num)
    else:
        set_active_backup(manager)

#Hacking pyplot !
Gcf.set_active = gcf_set_active

def restore_gcf_set_active():
    Gcf.set_active = set_active_backup
           
class PlotPanel(BasePanel):
    panelCategory = 'plot'
    panelShortName = 'basic'
    userVisible = True 
    
    classIconFile = str(respath / 'icons' / 'px16' / 'chart_curve.png')
    
    def __init__(self, parent=None, pid=None, figure_or_canvas=None):
        super().__init__(parent, pid, 'plot')
        
        self.fileMenu = self.menuBar().addMenu("&File")        
        self.addMenuItem(self.fileMenu, 'Open...' , self.openFigure)
        self.addMenuItem(self.fileMenu, 'Save As...' , self.saveFigure)
        self.addMenuItem(self.fileMenu, 'Close' , self.close_me_from_menu, icon = 'cross.png')
        
        self.viewMenu = CheckMenu("&View", self.menuBar())
        self.menuBar().addMenu(self.viewMenu)
        
        self.addMenuItem(self.viewMenu, 'Refresh', self.refresh,
            statusTip = "Refresh the plot",
            icon = 'update.png')
        self.addMenuItem(self.viewMenu, 'Grid', self.grid,
            statusTip = "Show/hide grid",
            icon = 'layer_grid.png')
        self.addMenuItem(self.viewMenu, 'Tight', self.tight)        
        self.addMenuItem(self.viewMenu, 'Interactive', self.toggle_interactive,
            checkcall=lambda: pylab.isinteractive())
        
        self.addBaseMenu()        
        
        if hasattr(figure_or_canvas, 'canvas'):
            self.figure = figure_or_canvas
            self.canvas = figure_or_canvas.canvas            
        else:
            self.figure = figure_or_canvas.figure
            self.canvas = figure_or_canvas                  
        
        self.setCentralWidget(self.canvas)         
        
        self.nav = NavigationToolbar(self.canvas, self)
        self.nav.setIconSize(QtCore.QSize(24,24))
        self.addToolBar(self.nav)        
        self.statusBar().hide()
        #self.statusBar().addWidget(self.nav)
        
    def showNewFigure(self, figure):     
        from ...matplotbe import FigureCanvasGh2    
        
        mgr = self.canvas.manager
        self.figure = figure
        self.canvas = FigureCanvasGh2(figure)
        self.canvas.manager = mgr
        mgr.canvas = self.canvas
        
        self.setCentralWidget(self.canvas)
        
        self.statusBar().removeWidget(self.nav)
        
        self.nav = NavigationToolbar(self.canvas, self)
        self.nav.setIconSize(QtCore.QSize(24,24))        
        
        #self.statusBar().addWidget(self.nav)
        self.addToolBar(self.nav) 
        
        self.canvas.draw_idle()   

    def openFigure(self):
        import pickle
        
        filter = "Pickle (*.pkl)"
        
        filepath, selectedFilter = gui.getfile(filter=filter, title='Open Figure')

        if filepath == '':
            return        
            
        pickle.load(open(filepath, 'rb'))

    def saveFigure(self):
        figure = self.figure

        filter = "Portable Network Graphics (*.png)"
        filter += ";;Scalable Vector Graphics (*.svg)"
        filter += ";;Pickle (*.pkl)"
        
        filepath, selectedFilter = gui.putfile(filter=filter, title='Save Figure as', defaultfilter="Scalable Vector Graphics (*.svg)")
        
        if filepath == '':
            return                
        
        if selectedFilter == "Pickle (*.pkl)":
            import pickle
            with open(filepath, 'wb') as fp:
                pickle.dump(figure, fp)
        else:
            plt = gui.prepareplot()
            plt.savefig(filepath)
        
    def grid(self):
        f = self.figure
        for axe in f.axes:
            axe.grid()
            
        self.canvas.draw_idle()
        
    def tight(self):
        self.figure.tight_layout()
        self.canvas.draw_idle()
        
    def toggle_interactive(self):
        pylab.interactive(not pylab.isinteractive())
        
    def refresh(self):
        self.canvas.draw_idle()
                
    def select(self):      
        manager = self.canvas.manager
        set_active_backup(manager)
        super().select()
        
    def close_me_from_menu(self):  
        #Use the matplotlib backend to close
        pylab.close(self.panid)
        
    def close_panel(self):              
        #Called from the matplotlib backend
        super().close_panel()

      
        
        


import collections
import importlib
import pprint
import logging
import pathlib

from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy import QtCore, QtGui, QtWidgets

from .. import gui, config

from ..panels.base import BasePanel

from .laystruct import LayoutStruct
from .docks import DockBase
from .boxes import DockHBox, DockVBox, DockBox

respath = pathlib.Path(config['respath'])
logger = logging.getLogger(__name__)

DEBUG = {'Use_ScrollBox': True}

class DockTabBar(QTabBar):
    def __init__(self, *args, **kwargs):        
        super().__init__(*args, **kwargs)                
        self.setMovable(True) 

        fontmetric = QtGui.QFontMetrics(self.font())
        self.fontheight = fontmetric.height()       

        self.movingWindow = None
        self.movingWindowOffset = None

    def add_tab_menu(self, index, leftWidget, rightWidget=None):                
        if index >= 0:
            if not leftWidget is None:
                self.setTabButton(index, QTabBar.LeftSide, leftWidget)            
            if not rightWidget is None:
                self.setTabButton(index, QTabBar.RightSide, rightWidget)

    def mouseDoubleClickEvent(self, event):
        ezm = gui.qapp.panels.ezm
        widget = self.parent().currentWidget()
        container = widget.get_container()        
        globalpos = widget.mapToGlobal(self.pos())
        
        if event.modifiers() == 0:            
            window, node = self.detach()
            if not window is container.parent():
                window.move(globalpos)
            return                

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        widget = self.parent().currentWidget()        
        self.drag_start_pos = event.pos()
        
        if isinstance(widget, BasePanel): 
            if event.button() == Qt.LeftButton:   
                widget.select()  

    def mouseReleaseEvent(self, event):        
        widget = self.parent().currentWidget()        
        
        if event.button() == Qt.RightButton:
            if not self.movingWindow is None:
                self.movingWindow.windowPlaced()
                self.movingWindow = None
                
            elif isinstance(widget, BasePanel):
                #Show the panel menu
                menu = QMenu('menu')
                for child in widget.menuBar().children():
                    #One of the QMenu children seems to be itself
                    if isinstance(child, QMenu) and child.title() != '':
                        menu.addMenu(child)                        
                menu.exec_(QCursor.pos())
        else:
            super().mouseReleaseEvent(event)    

    def mouseMoveEvent(self, event):
        movePoint = self.drag_start_pos - event.pos()
        moveDistance = movePoint.x() ** 2 + movePoint.y() ** 2
        if event.buttons() == Qt.RightButton and moveDistance > 32:
            if self.movingWindow is None:
                pos = self.mapToGlobal(QPoint(0, 0))
                self.movingWindowOffset = QtGui.QCursor.pos() - pos
                window, node = self.detach()
                window.startMoving()
                self.movingWindow = window
            else:
                self.movingWindow.move(QtGui.QCursor.pos() - self.movingWindowOffset)

        else:
            super().mouseMoveEvent(event)

    def detach(self):
        if isinstance(self.parent(), DockTag):
            widget = self.parent()
            
        elif isinstance(self.parent(), DockTab):
            widget = self.parent().currentWidget()
            
        if isinstance(widget, BasePanel):
            #It this case still used?
            logger.debug('Detaching BasePanel')            
            container = self.parent().get_container()
            geo = widget.geometry()
            window, node = container.detach('panel', widget.category, widget.panid, True, geo.width(), geo.height())
        else:
            logger.debug('Detaching Layout')
            window, node = widget.detach()
        
        return window, node        
        
class DockTabBase(QTabWidget, DockBase):

    def __init__(self, parent=None, collapse=None):
        QTabWidget.__init__(self, parent=parent)                 
        
        if DEBUG['Use_ScrollBox']:
            self.toprightbtn = QtWidgets.QToolButton(self)
            self.toprightbtn.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_split.png')))
            #self.toprightbtn.setText('...')
            self.toprightbtn.setPopupMode(QtWidgets.QToolButton.InstantPopup)
            self.setCornerWidget(self.toprightbtn, Qt.TopRightCorner)
            self.pinMenu = QtWidgets.QMenu('pin')            
            if config.get("scroll_area", False):
                self.pinMenu.addAction(QtWidgets.QAction('Move to Pinned/Scroll Area', self, triggered=self.moveToOtherArea))            
            self.pinMenu.addAction(QtWidgets.QAction('Duplicate', self, triggered=self.duplicate,
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_double.png'))))
            self.pinMenu.addAction(QtWidgets.QAction('Split Horizontal', self, triggered=self.splitHorizontal,
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'layouts_split.png'))))           
            self.pinMenu.addAction(QtWidgets.QAction('Split Vertical', self, triggered=self.splitVertical,
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'layouts_split_vertical.png'))))
            self.pinMenu.addSeparator()
            action = QtWidgets.QAction('Screenshot to Clipboard', self, triggered=self.screenShot)
            action.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'lcd_tv_image.png')))
            self.pinMenu.addAction(action)            
            self.pinMenu.addAction(QtWidgets.QAction('Toggle global Menu usage', self, triggered=self.toggleMenu,
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'menubar.png'))))
            self.pinMenu.addAction(QtWidgets.QAction('Show/Hide Status Bar', self, triggered=self.toggleStatusBar,
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'status_bar.png'))))
            self.toprightbtn.setMenu(self.pinMenu)
        
        if collapse == 'h':
            self.topleftbtn = QtWidgets.QToolButton(self)
            self.topleftbtn.setIcon(gui.qapp.resizeicon)
            self.topleftbtn.setCheckable(True)
            self.topleftbtn.setDisabled(True)
            #self.topleftbtn.clicked.connect(self.toggleHorizontalContent)
            self.setCornerWidget(self.topleftbtn, Qt.TopLeftCorner)
            
        elif collapse == 'v':
            self.topleftbtn = QtWidgets.QToolButton(self)
            self.topleftbtn.setIcon(gui.qapp.resizeicon)
            self.topleftbtn.setCheckable(True)
            self.topleftbtn.clicked.connect(lambda: self.parent().collapse(self))
            self.setCornerWidget(self.topleftbtn, Qt.TopLeftCorner)
            
        self.setTabBarAutoHide(False)  
        self.setTabBar(DockTabBar())        
        
        self.title = None   
        self.setAutoFillBackground(True)                
        
        self.collapsed = False
        
    def duplicate(self):
        panel = self.currentWidget()
        newpanel = panel.duplicate()
        newpanel.show_me()
        
    def splitHorizontal(self):
        panel = self.currentWidget()    
        container = panel.get_container()        
        newpanel = panel.duplicate(floating=True)        
        container.insert((newpanel.category, newpanel.panid), 'right', (panel.category, panel.panid))                
        return panel.panid
        
    def splitVertical(self):
        panel = self.currentWidget()
        container = panel.get_container()
        newpanel = panel.duplicate(floating=True)
        container.insert((newpanel.category, newpanel.panid), 'bottom', (panel.category, panel.panid))        
        return panel.panid
        
    def moveToOtherArea(self):
        self.get_dock_box().moveToOtherArea(self)
        
    def collapseVertical(self):
        if self.collapsed:
            for pos in range(self.count()):
                self.widget(pos).show()            
            self.setMinimumHeight(self.priorMinimumHeight)
            self.setMaximumHeight(self.priorMaximumHeight)
            try:
                self.parent().changeWidgetSize(self, self.priorHeight, fixed=False)
            except:
                pass
                
            self.collapsed = False
            self.topleftbtn.setChecked(False)
        else:
            for pos in range(self.count()):
                self.widget(pos).hide()
            self.priorMinimumHeight = self.minimumHeight()
            self.priorMaximumHeight = self.maximumHeight()
            self.priorHeight = self.height()
            tabbarheight = self.tabBar().height()
            try:
                self.parent().changeWidgetSize(self, tabbarheight, fixed=True)
                self.setFixedHeight(tabbarheight)
            except:
                self.setFixedHeight(tabbarheight)
            self.collapsed = True
            self.topleftbtn.setChecked(True)                     

    def set_tab_header(self, widget, title):
        index = self.indexOf(widget)
        
        bindButton = QToolButton(widget)
        bindButton.setIcon(gui.qapp.bindicon)
        #bindButton.setFixedHeight(26)
        bindButton.setCheckable(True)
        
        if widget.isSelected():            
            bindButton.setChecked(True)
        
        bindButton.clicked.connect(widget.select)
        bindButton.setPopupMode(QToolButton.MenuButtonPopup)
        bindButton.setMenu(widget.bindMenu)           
        bindButton.setFixedHeight(self.tabBar().fontheight + 4)
        
        gui.qapp.panels.ezm.add_button_to_bindgroup(widget.category, widget.panid, bindButton)
                
        self.tabBar().add_tab_menu(index, bindButton, None)
        
    def mouseReleaseEvent(self, event):
        if event.button() == Qt.RightButton:
            self.get_container().parent().newPanelMenu.exec_(QCursor.pos())
        super().mouseReleaseEvent(event)        
        
    def screenShot(self):      
        pixmap = self.currentWidget().grab()
        qimage = pixmap.toImage()
        clipboard = gui.qapp.clipboard()
        clipboard.setImage(qimage)        
        
    def toggleMenu(self):
        panel = self.currentWidget()
        panel.use_global_menu = not panel.use_global_menu
        panel.select()    

    def toggleStatusBar(self):
        panel = self.currentWidget()
        statusBar = panel.statusBar()
        if statusBar.isVisible():
            statusBar.hide()
        else:
            statusBar.show()           
        
        
class DockTab(DockTabBase):
    category = 'tab'
    
    def __init__(self, parent=None, collapse=None):
        super().__init__(parent, collapse)  
        self.setStyleSheet("QTabBar::tab { height: " +  str(self.tabBar().fontheight + 6) + ";}")        
        
    def addWidget(self, widget, title=None):
        if isinstance(widget, BasePanel):
            self.addTab(widget, title)       
            self.set_tab_header(widget, title)
        else:
            self.addTab(widget, title)

class DockTag(DockTabBase):
    category = 'tag'
    def __init__(self, parent=None, collapse=None):
        super().__init__(parent, collapse)

    def addWidget(self, widget, title=None):
        self.addTab(widget, title)

        if title.startswith('vbox'):
            self.setTabPosition(QTabWidget.West) 
            pal = self.palette()     
            pal.setColor(QPalette.Background, QColor(192,192,224))
            self.setPalette(pal)

        elif title.startswith('hbox'):
            pal = self.palette()     
            pal.setColor(QPalette.Background, QColor(192,224,192))
            self.setPalette(pal)

        else:
            self.setStyleSheet("QTabBar::tab { height: " +  str(self.tabBar().fontheight + 6) + ";}")
            pal = self.palette()     
            pal.setColor(QPalette.Background, QColor(242,242,242))
            self.setPalette(pal)     

        if isinstance(widget, BasePanel):    
            self.addTab(widget, title)       
            self.set_tab_header(widget, title)     
        else:
            self.addTab(widget, title)    


class DockContainer(QWidget):
    def __init__(self, manager, parent=None, name='main'):
        super().__init__(parent=parent)        
        
        #QWidget needs a top level QLayout object
        #Note that a QSplitter is not a QLayout object
        
        layout = QVBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(0)
        self.taglevel = None
        
        self.setLayout(layout)   
        
        self.manager = manager
        self.name = name
        self.laywidget = None
        
        self.panelIds = []
        
        self.tabindex = 0
        self.vboxindex = 0
        self.hboxindex = 0 

    @property
    def all_panels(self):
        return self.manager.panels

    def is_empty(self, check=False):
        if check:
            return self.get_layout_struct().is_empty()
        else:
            return len(self.panelIds) == 0
            
    def panel_count(self):
        return len(self.panelIds)

    def update_layout(self, layout_struct):    
        if not self.laywidget is None:
            self.detach_panels()
            self.layout().removeWidget(self.laywidget)               
            
        self.tabindex = 0
        self.vboxindex = 0
        self.hboxindex = 0
        parentnode = {'type': 'layout', 'category': 'root'}                        
        
        if self.taglevel is None:
            if layout_struct.root.get('type', None) == 'panel' and config.get("hide_solo_panel", False):
                self.taglevel = 0
            else:
                self.taglevel = 1
            
        self.laywidget = self.make_layout_widget_branch(layout_struct.root, parentnode)
        self.laywidget.show()        
        #self.show_all(self.laywidget)
        
        self.layout().addWidget(self.laywidget)        

    def detach_panels(self):        
        for cat, panid in self.panelIds:
            try:
                panel = self.all_panels[cat][panid]
                panel.detach()                
            except KeyError:
                print(f'Could not find panel {cat}#{panid}')                
        self.panelIds.clear()  

    def detach_top(self):
        l = self.get_layout_struct()
        return self.detach(l.root['type'], l.root['category'], l.root['id'], False)

    def detach(self, nodetype, category, nodeid, to_new_window=True, width=640, height=480):                
        drop_layout = self.get_layout_struct()         
        place_layout = drop_layout.pop_node(nodetype, category, nodeid)  #tag or tab?
        
        drop_layout.compact()    
        place_layout.compact()
        
        if drop_layout.is_empty():
            return self.parent(), place_layout        
        
        self.update_layout(drop_layout)                           
        window = self.manager.new_window_using_layout(place_layout, width, height, self.parent().name)
        
        return window, place_layout

    def compact(self):
        ls = self.get_layout_struct()
        ls.compact()
        self.update_layout(ls)

    def distribute(self):
        ls = self.get_layout_struct()
        ls.compact()
        ls.distribute()
        self.update_layout(ls)
        
    def insert(self, panel, relative_pos, to_panel):
        """
        :param dict insert_node: (category, panid)
        :param str relative_pos: 'tab', 'top', 'bottom', 'left' or 'right'
        :param refpanqualid: (category, panid)
        """        
        ls = self.get_layout_struct()
        ls.compact()
        ls.insert_panel(panel, relative_pos, to_panel)   
        self.update_layout(LayoutStruct())
        self.update_layout(ls)        

    def make_layout_widget_branch(self, node, parentnode=None, ind=None):
        if len(node.keys()) == 0:
            lay = DockTab()
            return lay
            
        elif node['type'] == 'panel': 
            #No layout required at all
            cat = node['category']
            panid = node['id']
            panel = self.all_panels[cat][panid]
            panel.title = panel.short_title            
            self.panelIds.append((cat, panid))
            lay = panel
            
        elif node['type'] == 'layout':
            if node['category'] == 'hbox':
                self.hboxindex += 1
                lay = DockHBox()
                lay.title = f'hbox#{self.hboxindex}'
                
            elif node['category'] == 'vbox':
                self.vboxindex += 1
                lay = DockVBox()
                lay.title = f'vbox#{self.vboxindex}'

            elif node['category'] == 'tab':            
                self.tabindex += 1
                if parentnode['category'] == 'hbox':
                    lay = DockTab(None, 'h')
                elif parentnode['category'] == 'vbox':
                    lay = DockTab(None, 'v')
                else:
                    lay = DockTab()
                lay.title = f'tab#{self.tabindex}'                         
                
            else:
                raise TypeError(f'Unknown node category {node}')
        else:
            raise TypeError(f'Unknown node type {node}')
            
        lay.nodeinfo = {'parent': parentnode, 'index': ind}

        if isinstance(lay, DockBox):
            items = node.get('items', [])
            pin_sizes = node.get('sizes', [])
            scroll_sizes = node.get('scroll', [])
            areas = len(pin_sizes) * [DockBox.PinArea] + len(scroll_sizes) * [DockBox.ScrollArea]
            for ind, (item, area) in enumerate(zip(items, areas)):
                branch = self.make_layout_widget_branch(item, node, ind)
                lay.addWidget(branch, area=area, title=branch.title)            
        else:
            for ind, item in enumerate(node.get('items', [])):
                branch = self.make_layout_widget_branch(item, node, ind)
                lay.addWidget(branch, title=branch.title)

        if node['type'] == 'layout':
            if 'sizes' in node.keys():
                lay.setSizes(node['sizes'])
                
            if 'scroll' in node.keys():
                lay.setSizes(node['scroll'], area=DockBox.ScrollArea)    

            if 'pinscroll' in node.keys():
                lay.setSizes(node['pinscroll'], area=DockBox.SplitArea)                
                
            if 'active' in node.keys():               
                lay.setCurrentIndex(node['active'])
            
        if parentnode['category'] == 'tab':
            return lay
            
        if (node['type'] == 'panel' and self.taglevel in [0,1]) or self.taglevel == 2:
            if parentnode['category'] == 'hbox':
                tag = DockTag(None, 'h')
            elif parentnode['category'] == 'vbox':
                tag = DockTag(None, 'v')
            else:
                tag = DockTag()
                
            if self.taglevel == 0:
                tag.setTabBarAutoHide(True)
                
            tag.title = lay.title
            tag.addWidget(lay, title=lay.title)
            lay = tag
                
        return lay    

    def get_layout_struct(self):
        layout_widget = self.laywidget
        
        if not layout_widget is None and layout_widget.parent() == self:
            root = DockContainer.from_layout_widget_branch(layout_widget)
            ls = LayoutStruct()
            ls.root = root            
        else:
            ls = LayoutStruct()
        return ls                   
    
    @staticmethod
    def from_layout_widget_branch(layout_widget):
        node = {}
        
        if isinstance(layout_widget, (DockHBox, DockVBox)):
            if layout_widget.orientation() == Qt.Orientation.Horizontal:
                node['type'] = 'layout'
                node['category'] = 'hbox'
                node['id'] = layout_widget.title
            else:
                node['type'] = 'layout'
                node['category'] = 'vbox'
                node['id'] = layout_widget.title
                
            node['sizes'] = layout_widget.sizes()
            
            try:
                node['pinscroll'] = layout_widget.sizes(DockBox.SplitArea)
                node['scroll'] = layout_widget.sizes(DockBox.ScrollArea)
            except:
                pass
                
        elif isinstance(layout_widget, DockTab):
            node['type'] = 'layout'
            node['category'] = 'tab'
            node['id'] = layout_widget.title
            node['active'] = layout_widget.currentIndex()
                
        elif isinstance(layout_widget, DockTag):
            node['type'] = 'layout'
            node['category'] = 'tag'
            node['id'] = layout_widget.title            
                
        else:
            panel = layout_widget
            node = {'type': 'panel', 'category': panel.category, 'id': panel.panid}
            return node
            
        node['items'] = []
        
        if isinstance(layout_widget, DockBox):        
            for index in range(layout_widget.count(DockBox.PinArea)):
                widget = layout_widget.widget(index, DockBox.PinArea)
                subnode = DockContainer.from_layout_widget_branch(widget)
                node['items'].append(subnode)     

            for index in range(layout_widget.count(DockBox.ScrollArea)):
                widget = layout_widget.widget(index, DockBox.ScrollArea)
                subnode = DockContainer.from_layout_widget_branch(widget)
                node['items'].append(subnode) 

        else:
            for index in range(layout_widget.count()):
                widget = layout_widget.widget(index)
                subnode = DockContainer.from_layout_widget_branch(widget)
                node['items'].append(subnode)                 
            
        return node     

    def show_all(self, layout_widget):
        if not (isinstance(layout_widget, QSplitter) or isinstance(layout_widget, (DockTab, DockTag))):
            layout_widget.show()
            
        else:        
            for index in range(layout_widget.count()):
                widget = layout_widget.widget(index)
                self.show_all(widget)
                
    def cycle_tag_level(self):
        if len(self.panelIds) < 2:
            self.taglevel = (self.taglevel + 1) % 2
        else:
            self.taglevel = (self.taglevel + 1) % 3
        self.compact()        



# -*- coding: utf-8 -*-
from PySide2 import QtGui, QtCore, QtXml


class Node(object):

    def __init__(self, name, parent=None):
        self._name = name
        self._parent = parent
        self._children = []
        self._value = None
        if parent is not None:
            parent.addChild(self)

    def typeInfo(self):
        return 'NODE'

    def addChild(self, child):
        self._children.append(child)

    def insertChild(self, position, child):
        if position < 0 or position > len(self._children):
            return False

        self._children.insert(position, child)
        child._parent = self
        return True

    def removeChild(self, position):
        if position < 0 or position > len(self._children):
            return False

        self._children.pop(position)
        child._parent = None
        return True

    def attrs(self):
        classes = self.__class__.__mro__
        keyvalued = {}
        for cls in classes:
            for key, value in cls.__dict__.iteritems():
                if isinstance(value, property):
                    keyvalued[key] = value.fget(self)
        return keyvalued

    def to_xml(self):
        doc = QtXml.QDomDocument()
        node = doc.createElement(self.typeInfo())
        doc.appendChild(node)

        for i in self._children:
            i._recurseXml(doc, node)

        return doc.toString(indent=4)

    def _recurse_xml(self, doc, parent):
        node = doc.createElement(self.typeInfo())
        parent.appendChild(node)
        attrs = self.attrs().iteritems()

        for k, v in attrs:
            node.setAttribute(k, v)

        for child in self._children:
            child._recurse_xml(doc, node)

    def to_list(self):
        output = []
        if self._children:
            for child in self._children:
                output += [self.name, child.to_list()]
        else:
            output += [self.name, self.value]
        return output

    def to_dict(self, d={}):
        for child in self._children:
            child._recurse_dict(d)
        return d

    def _recurse_dict(self, d):
        if self._children:
            d[self.name] = {}
            for child in self._children:
                child._recurse_dict(d[self.name])
        else:
            d[self.name] = self.value

    def name():
        def fget(self):
            return self._name
        def fset(self, value):
            self._name = value
        return locals()
    name = property(**name())

    def value():
        def fget(self):
            return self._value
        def fset(self, value):
            self._value = value
        return locals()
    value = property(**value())

    def child(self, row):
        return self._children[row]

    def childCount(self):
        return len(self._children)

    def parent(self):
        return self._parent

    def row(self):
        if self._parent is not None:
            return self._parent._children.index(self)

    def log(self, tabLevel=-1):
        output = ''
        tabLevel += 1

        for i in range(tabLevel):
            output += '    '

        output += ''.join(('|----', self._name,' = ', '\n'))

        for child in self._children:
            output += child.log(tabLevel)

        tabLevel -= 1
        output += '\n'
        return output

    def __repr__(self):
        return self.log()

    def data(self, column):
        if   column == 0:
            return self.name
        elif column == 1:
            return self.value

    def setData(self, column, value):
        if column == 0:
            self.name = value
        if column == 1:
            self.value = value

    def resource(self):
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Build an installable package."""

import io
import os
import sys
import glob
from pathlib import Path

from setuptools import find_packages, setup, Command, Distribution

herepath = Path(__file__).parent.absolute()
here = str(herepath)

MODULE_NAME = 'gdesk'
DISTRO_NAME = 'gamma-desk'
DESCRIPTION = 'A Python work environment image viewers & plots'
URL = 'https://github.com/thocoo/gamma-desk'
EMAIL = 'thomas.cools@telenet.be'
AUTHOR = 'Thomas Cools'

modpath = herepath / 'gdesk'

REQUIRED = [
    'numpy', 
    'imageio',
    'matplotlib',
    'scipy',
    'PySide2',
    'qtpy',
    'psutil',
    'numba',
    'pyzmq',
    'ninja2',
    'toml'
]

PYTHON_REQUIRED = '>=3.6'

def get_resources():
    found_resources = []    

    found_resources.append(str(modpath / 'config' / 'defaults.json'))

    for path in modpath.glob('resources/**/*'):
        if path.is_dir(): continue
        found_resources.append(str(path))

    return found_resources

with open(modpath / 'version.py') as fp:
    exec(fp.read())

# Import the README and use it as the long-description.    
with open(herepath / 'README.md', encoding='utf-8') as fp:
    LONG_DESCRIPTION = '\n' + fp.read()

setup(
    name=DISTRO_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    license='Apache License 2.0',
    url=URL,
    packages=find_packages(exclude=('tests',)),
    package_data=dict(gdesk=get_resources(),),
    entry_points={'console_scripts': [f'{MODULE_NAME} = {MODULE_NAME}.console:argexec']},
    install_requires=REQUIRED,
    include_package_data=True,
    python_requires=PYTHON_REQUIRED,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',        
    ],
)



# This module contains a script manager which implements Matlab like live scripting
# Support call of scripts which can be automatic reloaded without leaving the Python interpreter
# The user can modify code in a script file
# These script files will be reloaded based on modify time stamps
# New compiled code is direct available in current Python process

import os, sys, traceback, time
from pathlib import Path

def show_syntax_error(writer_call):
    """Display the syntax error that just occurred."""
    type, value, tb = sys.exc_info()
    sys.last_type = type
    sys.last_value = value
    sys.last_traceback = tb

    lines = traceback.format_exception_only(type, value)
    writer_call(''.join(lines))

def show_traceback(writer_call):
    """Display the exception that just occurred."""
    try:
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        tblist = traceback.extract_tb(tb)
        del tblist[:1]
        lines = traceback.format_list(tblist)
        if lines:
            lines.insert(0, "Traceback (most recent call last):\n")
        lines.extend(traceback.format_exception_only(type, value))
    finally:
        tblist = tb = None

    writer_call(''.join(lines))
    
class MarkUpdateRun(object):
    def __init__(self, scm, ls_code, attr):
        self.scm = scm
        self.ls_code = ls_code
        self.attr = attr
        
    @property
    def func(self):
        return getattr(self.ls_code.workspace, self.attr)
        
    def __call__(self, *args, **kwargs):
        self.scm.mark_for_update()
        ls_code = self.ls_code        
        error = ls_code.check_for_update()                        
        return self.func(*args, **kwargs)
        
    @property
    def __doc__(self):
        return self.func.__doc__
        
    def __repr__(self):
        return f'<Mark, Update and Run Callable {self.attr}>'        

class LiveScriptModule(object):
    def __init__(self, script_manager, path, top=False):
        object.__setattr__(self, '__script_manager__', script_manager)
        object.__setattr__(self, '__path__', path)
        object.__setattr__(self, '__top__', top)
        
    @property
    def __wrapped__(self):
        ls_code = self.__script_manager__.ls_codes[self.__path__]       
        error = ls_code.check_for_update()            
        return ls_code.workspace

    def __getattr__(self, attr):            
        if self.__top__:
            wrapped_attr = getattr(self.__wrapped__, attr)
            if callable(wrapped_attr):
                return MarkUpdateRun(self.__script_manager__, self.__script_manager__.ls_codes[self.__path__], attr)
        else:
            wrapped_attr = getattr(self.__wrapped__, attr)
        return wrapped_attr

    def __setattr__(self, attr, value):
        setattr(self.__wrapped__, attr, value)

    def __dir__(self):
        if self.__top__:
            self.__script_manager__.mark_for_update()
        return self.__wrapped__.__dir__()

    def __repr__(self):
        return f"<LiveScriptModule '{self.__path__}'>"

    @property
    def __doc__(self):
        return self.__wrapped__.__doc__
        
class LiveScriptTree(object):        
    def __init__(self, script_manager, path, top=False, name=None):
        object.__setattr__(self, '__script_manager__', script_manager)
        object.__setattr__(self, '__path__', path)
        object.__setattr__(self, '__top__', top)
        object.__setattr__(self, '__name__', name)

    def __dir__(self):
        lst = []
        lst = list(self.__dict__.keys())
        lst.extend(list(type(self).__dict__.keys()))        
        node = self.__path__
        for file in node.glob('*'):
            lst.append(file.stem)
        return lst
        
    def __getattr__(self, attr):
        path = self.__path__ / attr
        qualname = f'{self.__name__}.{attr}'
        return self.__script_manager__.using_path(path, top=self.__top__, name=qualname)        
            
    def __repr__(self):
        return f"<LiveScriptTree '{self.__path__}'>"            
        
class LiveScriptScan(object):
    def __init__(self, script_manager, top=False):
        object.__setattr__(self, '__script_manager__', script_manager)
        object.__setattr__(self, '__top__', top)
        object.__setattr__(self, '__name__', 'LiveScriptScan')

    def __dir__(self):
        lst = []
        lst = list(self.__dict__.keys())
        lst.extend(list(type(self).__dict__.keys()))        
        paths = self.__script_manager__.path
        for path in paths:            
            node = Path(path)
            for file in node.glob('*'):
                if file.is_dir():
                    lst.append(file.stem)
                elif file.suffix.lower() == '.py':
                    lst.append(file.stem)
                    
        return lst
        
    def __getattr__(self, attr):
        try:
            return self.__using__(attr)
        except KeyError:
            #Without this, Jypeter doesn't auto-complete
            #I don't know why
            raise AttributeError(attr)      
        

    def __call__(self, modstr):
        return self.__using__(modstr)
    
    def __using__(self, modstr):    
        self.__script_manager__.log(f'Calling {modstr}')
        
        if isinstance(self.__top__, str):
            top = sys._getframe(2).f_globals['__name__'] == self.__top__
        else:
            top = self.__top__
            
        return self.__script_manager__.using_modstr(modstr, top)
            
    def __repr__(self):
        return f"<LiveScriptScan '{self.__script_manager__.path}'>"       
            
            
class LsCode(object):        
    def __init__(self, script_manager, path, name=None):
        self.script_manager = script_manager
        self.path = path
        self.name = 'unknown' if name is None else name
        self.load_modify = -1
        self.code = None        
        self.workspace = None
        self.ask_refresh = 0        
        
    def check_for_update(self):
        ls_code = self
        updated = None
        if not ls_code.ask_refresh == 0:
            if ls_code.ask_refresh == 1:                
                updated = ls_code.update()                
                
            elif ls_code.ask_refresh == 2:
                updated = ls_code.load()                
                
            if not updated is None:
                if updated == 0:
                    ls_code.ask_refresh = 0
                    self.script_manager.log(f'Updated {ls_code.path}')    
                else:
                    self.script_manager.warn(f'Failed to update {ls_code.path}')
                    self.script_manager.warn(f'Error code {updated}')
                    
        return updated
        
    def modify_time(self):
        return os.path.getmtime(str(self.path))                
        
    def update(self):
        if self.load_modify < self.modify_time():
            ret = self.load()
            return ret
        else:
            return None            
        
    def load(self):
        self.code = None        
        self.workspace = None
            
        with open(str(self.path), 'r', encoding='utf-8') as fp:
            current_modify_stamp = self.modify_time()
            pycode = fp.read()            
            
            try:
                codeobj = compile(pycode, str(self.path), 'exec')
                self.code = codeobj
            except:
                show_syntax_error(self.script_manager.write_syntax_err)
                return 1
                
        self.workspace = LsWorkspace(self, str(self.path), self.name) 
                
        try:
            exec(codeobj, self.workspace.__dict__, self.workspace.__dict__)
                    
        except:
            show_traceback(self.script_manager.write_error)
            return 2
            
        self.load_modify =  current_modify_stamp       
        self.script_manager.log(f'{self.path} loaded at {time.ctime(self.load_modify)}')        
        return 0 
        
        
class LsWorkspace(object):
    #Provide a namespace for each script file.
    def __init__(self, ls_code, file, name='unkown'):
        self.__file__ = file
        self.__ls_code__ = ls_code
        self.__name__ = name
        
class LiveScriptManager(object):
    def __init__(self, workspace=None):
        # Add the search paths to self.path
        if workspace is None:
            workspace = dict()
        self.path = []        
        self.ls_codes = dict()
        self.workspace = workspace
        self.verbose = 3
        
    def find_script(self, modstr='test'):
        """Search for the script in the path.
        Return the found path.
        """
        modpath = modstr.replace('.', '/')
        
        result = []
        
        for path in self.path:
            #It is note sure every path exists
            path = Path(path).absolute()
            if (path / modpath).with_suffix('.py').exists():
                result.append(((path / modpath).with_suffix('.py'), 'file'))
            elif (path / modpath).is_dir():
                result.append(((path / modpath), 'dir'))
            
        if len(result) == 0:
            raise KeyError(f'{modstr} not found')
        elif len(result) == 1:           
            return result[0]
        else:
            self.warn(f'Multiple matches found for {modstr}')
            for path in result:
                self.warn(str(path[0]))
            return result[0]
                
    def append_path(self, path, resolve=True):
        path = Path(path).absolute()
        if resolve:
            try:
                path = path.resolve()
            except:
                print(f'Script path {path} not found')
                path = None
                
        if not path is None and not path in self.path:
            self.path.append(str(path))
        
    def load(self, path, name=None):
        self.ls_codes[str(path)] = LsCode(self, path, name)
        self.ls_codes[str(path)].load()
        
    def log(self, *msgs):
        if self.verbose > 4:
            print('<LiveScriptManager Log', *msgs, '>')        
            
    def warn(self, *msgs):
        if self.verbose > 2:
            print('<LiveScriptManager Warning', *msgs, '>')             
        
    def update_now(self, enforce=False):
        """
        Reload the scripts in memory.
        If not enforced, load only scripts with more recent timestamps
        """
        for ls_code in self.ls_codes.values():
            if enforce:
                ret = ls_code.load()
            else:
                ret = ls_code.update()               
                
    def mark_for_update(self, enforce=False):
        self.log('Marking all modules for update')
        mark = 2 if enforce else 1
        for ls_code in self.ls_codes.values():
            ls_code.ask_refresh = mark           
        
    def using_modstr(self, modstr, top=False):
        """Load a script or make a ScriptTree.
        A ScripTree is used to link to a dir.
        The loading of a script is done at moment of attribute access.
        """           
        path, stype = self.find_script(modstr)
        return self.using_path(path, stype, top, modstr)
        
    def using_path(self, path, stype=None, top=False, name=None):        
        if stype is None:
            if path.is_dir():
                stype = 'dir'                
            else:
                path = path.with_suffix('.py')
                stype = 'file'
                
        if path is None:
            raise ImportError(f'LiveScript {modstr} not found')
            
        if str(path) in self.ls_codes.keys():
            return LiveScriptModule(self, str(path), top)            
            
        if stype == 'file':
            self.load(path, name)            
            return LiveScriptModule(self, str(path), top)            
            
        elif stype == 'dir':
            return LiveScriptTree(self, path, top, name)                  

    def write_error(self, text):
        sys.stderr.write(text)       

    def write_syntax_err(self, text):
        sys.stderr.write(text)
        






"""
Edit on Jan 02, 2020
@author: the_RR
Adapted for python 3.4+

Requires pywin32
original: http://code.activestate.com/recipes/474121/
    # HtmlClipboard
    # An interface to the "HTML Format" clipboard data format

    __author__ = "Phillip Piper (jppx1[at]bigfoot.com)"
    __date__ = "2006-02-21"
    __version__ = "0.1"
"""

import re
import time
import random
import win32clipboard

#---------------------------------------------------------------------------
#  Convenience functions to do the most common operation

def HasHtml():
    """
    Return True if there is a Html fragment in the clipboard..
    """
    cb = HtmlClipboard()
    return cb.HasHtmlFormat()


def GetHtml():
    """
    Return the Html fragment from the clipboard or None if there is no Html in the clipboard.
    """
    cb = HtmlClipboard()
    if cb.HasHtmlFormat():
        return cb.GetFragment()
    else:
        return None


def PutHtml(fragment):
    """
    Put the given fragment into the clipboard.
    Convenience function to do the most common operation
    """
    cb = HtmlClipboard()
    cb.PutFragment(fragment)


#---------------------------------------------------------------------------

class HtmlClipboard:

    CF_HTML = None

    MARKER_BLOCK_OUTPUT = \
        "Version:1.0\r\n" \
        "StartHTML:%09d\r\n" \
        "EndHTML:%09d\r\n" \
        "StartFragment:%09d\r\n" \
        "EndFragment:%09d\r\n" \
        "StartSelection:%09d\r\n" \
        "EndSelection:%09d\r\n" \
        "SourceURL:%s\r\n"

    MARKER_BLOCK_EX = \
        "Version:(\S+)\s+" \
        "StartHTML:(\d+)\s+" \
        "EndHTML:(\d+)\s+" \
        "StartFragment:(\d+)\s+" \
        "EndFragment:(\d+)\s+" \
        "StartSelection:(\d+)\s+" \
        "EndSelection:(\d+)\s+" \
        "SourceURL:(\S+)"
    MARKER_BLOCK_EX_RE = re.compile(MARKER_BLOCK_EX)

    MARKER_BLOCK = \
        "Version:(\S+)\s+" \
        "StartHTML:(\d+)\s+" \
        "EndHTML:(\d+)\s+" \
        "StartFragment:(\d+)\s+" \
        "EndFragment:(\d+)\s+" \
           "SourceURL:(\S+)"
    MARKER_BLOCK_RE = re.compile(MARKER_BLOCK)

    DEFAULT_HTML_BODY = \
        "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">" \
        "<HTML><HEAD></HEAD><BODY><!--StartFragment-->%s<!--EndFragment--></BODY></HTML>"

    def __init__(self):
        self.html = None
        self.fragment = None
        self.selection = None
        self.source = None
        self.htmlClipboardVersion = None


    def GetCfHtml(self):
        """
        Return the FORMATID of the HTML format
        """
        if self.CF_HTML is None:
            self.CF_HTML = win32clipboard.RegisterClipboardFormat("HTML Format")

        return self.CF_HTML


    def GetAvailableFormats(self):
        """
        Return a possibly empty list of formats available on the clipboard
        """
        formats = []
        try:
            win32clipboard.OpenClipboard(0)
            cf = win32clipboard.EnumClipboardFormats(0)
            while (cf != 0):
                formats.append(cf)
                cf = win32clipboard.EnumClipboardFormats(cf)
        finally:
            win32clipboard.CloseClipboard()

        return formats


    def HasHtmlFormat(self):
        """
        Return a boolean indicating if the clipboard has data in HTML format
        """
        return (self.GetCfHtml() in self.GetAvailableFormats())


    def GetFromClipboard(self):
        """
        Read and decode the HTML from the clipboard
        """

        # implement fix from: http://teachthe.net/?p=1137

        cbOpened = False
        while not cbOpened:
            try:
                win32clipboard.OpenClipboard(0)
                src = win32clipboard.GetClipboardData(self.GetCfHtml())
                src = src.decode("UTF-8")
                #print(src)
                self.DecodeClipboardSource(src)

                cbOpened = True

                win32clipboard.CloseClipboard()
            except Exception as err:
                # If access is denied, that means that the clipboard is in use.
                # Keep trying until it's available.
                if err.winerror == 5:  # Access Denied
                    pass
                    # wait on clipboard because something else has it. we're waiting a
                    # random amount of time before we try again so we don't collide again
                    time.sleep( random.random()/50 )
                elif err.winerror == 1418:  # doesn't have board open
                    pass
                elif err.winerror == 0:  # open failure
                    pass
                else:
                    print( 'ERROR in Clipboard section of readcomments: %s' % err)

                    pass

    def DecodeClipboardSource(self, src):
        """
        Decode the given string to figure out the details of the HTML that's on the string
        """
        # Try the extended format first (which has an explicit selection)
        matches = self.MARKER_BLOCK_EX_RE.match(src)
        if matches:
            self.prefix = matches.group(0)
            self.htmlClipboardVersion = matches.group(1)
            self.html = src[int(matches.group(2)):int(matches.group(3))]
            self.fragment = src[int(matches.group(4)):int(matches.group(5))]
            self.selection = src[int(matches.group(6)):int(matches.group(7))]
            self.source = matches.group(8)
        else:
            # Failing that, try the version without a selection
            matches = self.MARKER_BLOCK_RE.match(src)
            if matches:
                self.prefix = matches.group(0)
                self.htmlClipboardVersion = matches.group(1)
                self.html = src[int(matches.group(2)):int(matches.group(3))]
                self.fragment = src[int(matches.group(4)):int(matches.group(5))]
                self.source = matches.group(6)
                self.selection = self.fragment


    def GetHtml(self, refresh=False):
        """
        Return the entire Html document
        """
        if not self.html or refresh:
            self.GetFromClipboard()
        return self.html


    def GetFragment(self, refresh=False):
        """
        Return the Html fragment. A fragment is well-formated HTML enclosing the selected text
        """
        if not self.fragment or refresh:
            self.GetFromClipboard()
        return self.fragment


    def GetSelection(self, refresh=False):
        """
        Return the part of the HTML that was selected. It might not be well-formed.
        """
        if not self.selection or refresh:
            self.GetFromClipboard()
        return self.selection


    def GetSource(self, refresh=False):
        """
        Return the URL of the source of this HTML
        """
        if not self.selection or refresh:
            self.GetFromClipboard()
        return self.source


    def PutFragment(self, fragment, selection=None, html=None, source=None):
        """
        Put the given well-formed fragment of Html into the clipboard.
        selection, if given, must be a literal string within fragment.
        html, if given, must be a well-formed Html document that textually
        contains fragment and its required markers.
        """
        if selection is None:
            selection = fragment
        if html is None:
            html = self.DEFAULT_HTML_BODY % fragment
        if source is None:
            source = "file://HtmlClipboard.py"

        fragmentStart = html.index(fragment)
        fragmentEnd = fragmentStart + len(fragment)
        selectionStart = html.index(selection)
        selectionEnd = selectionStart + len(selection)
        self.PutToClipboard(html, fragmentStart, fragmentEnd, selectionStart, selectionEnd, source)


    def PutToClipboard(self, html, fragmentStart, fragmentEnd, selectionStart, selectionEnd, source="None"):
        """
        Replace the Clipboard contents with the given html information.
        """

        try:
            win32clipboard.OpenClipboard(0)
            win32clipboard.EmptyClipboard()
            src = self.EncodeClipboardSource(html, fragmentStart, fragmentEnd, selectionStart, selectionEnd, source)
            src = src.encode("UTF-8")
            #print(src)
            win32clipboard.SetClipboardData(self.GetCfHtml(), src)
        finally:
            win32clipboard.CloseClipboard()


    def EncodeClipboardSource(self, html, fragmentStart, fragmentEnd, selectionStart, selectionEnd, source):
        """
        Join all our bits of information into a string formatted as per the HTML format specs.
        """
        # How long is the prefix going to be?
        dummyPrefix = self.MARKER_BLOCK_OUTPUT % (0, 0, 0, 0, 0, 0, source)
        lenPrefix = len(dummyPrefix)

        prefix = self.MARKER_BLOCK_OUTPUT % (lenPrefix, len(html)+lenPrefix,
                        fragmentStart+lenPrefix, fragmentEnd+lenPrefix,
                        selectionStart+lenPrefix, selectionEnd+lenPrefix,
                        source)
        return (prefix + html)


def DumpHtml():

    cb = HtmlClipboard()
    print("GetAvailableFormats()=%s" % str(cb.GetAvailableFormats()))
    print("HasHtmlFormat()=%s" % str(cb.HasHtmlFormat()))
    if cb.HasHtmlFormat():
        cb.GetFromClipboard()
        print("prefix=>>>%s<<<END" % cb.prefix)
        print("htmlClipboardVersion=>>>%s<<<END" % cb.htmlClipboardVersion)
        print("GetSelection()=>>>%s<<<END" % cb.GetSelection())
        print("GetFragment()=>>>%s<<<END" % cb.GetFragment())
        print("GetHtml()=>>>%s<<<END" % cb.GetHtml())
        print("GetSource()=>>>%s<<<END" % cb.GetSource())


if __name__ == '__main__':

    def test_SimpleGetPutHtml():
        data = "<p>Writing to the clipboard is <strong>easy</strong> with this code.</p>"
        PutHtml(data)
        if GetHtml() == data:
            print("passed")
        else:
            print("failed")

    def test_tabulated_data():
        import sys
        sys.path.append(r'E:\ProjectsClio\is_common\tools\utrm\pypack')

        from utrm.prettytable import PrettyTable

        tbl = PrettyTable(['Name','Value'])
        tbl.add_row(['register0', 1])
        tbl.add_row(['register1', 2])

        PutHtml(tbl.get_html_string())


    #test_SimpleGetPutHtml()
    test_tabulated_data()
    #DumpHtml()


import sys
import threading
from multiprocessing import Queue
import functools
import importlib
import types
import pickle
import logging
import pathlib
import time

import numpy as np

from ..utils.funccom import find_nested_func
from .conf import config

try:
    #Only on Windows    
    from ..utils.keypress import PressKey, ReleaseKey
    
except:
    pass

logger = logging.getLogger(__name__)

if config.get('sphinx', False):
    #https://stackoverflow.com/questions/28366818/preserve-default-arguments-of-wrapped-decorated-python-function-in-sphinx-docume
    def wraps(original_func):
       wrap_decorator = functools.wraps(original_func)
       def re_wrapper(func):
           wrapper = wrap_decorator(func)
           poorman_sig = original_func.__code__.co_varnames[
                             :original_func.__code__.co_argcount]
           wrapper.__doc__ = "{} ({})\n\n{}".format (
                original_func.__name__, ", ".join(poorman_sig),
                wrapper.__doc__) 
           return wrapper
       return re_wrapper           
else:           
    wraps = functools.wraps
    
def StaticGuiCall(func):
    #Decorator for pushing the function call through a queue
    #func is the function of the decorated method, not the method itself?    
    @staticmethod
    @wraps(func)
    def caller(*args, **kwargs):        
        return gui.gui_call(func, *args, **kwargs)
    return caller
    
    
class RedBull(object):
    def __init__(self, interval=60):       
        self.interval = interval
        self.actives = []
        self.timer_thread = None
        #Function Key 15
        self.keycode = 0x7E  

    def enable(self, interval=None):        
        if len(self.actives) == 0:
            self.start(interval or self.interval)
        else:
            self.stop() 
            self.start(interval or self.interval)
                                                               
    def disable(self):
        self.stop()
            
    def start(self, interval):
        self.timer_thread = threading.Thread(target=self.loop, args=(interval,))
        self.timer_thread.start()
        
    def stop(self):
        self.actives.clear()
        self.timer_thread = None
            
    def loop(self, interval):
        ident = threading.get_ident()
        self.actives.append(ident)
        while True:
            time.sleep(interval)
            if not ident in self.actives: break
            self.do()
        
    def do(self):
        PressKey(self.keycode)  
        ReleaseKey(self.keycode)       
    
        
class GuiProxyBase(object):
    category = None
    
    @classmethod
    def derivedClasses(cls):
        l = []
        for SubClass in cls.__subclasses__():
            l.extend(SubClass.derivedClasses())
            l.append((SubClass.category, SubClass))
            
        if cls is GuiProxyBase:
            result = dict()
            
            for category, Cls in l:
                if not category in result.keys():
                    result[category] = []
                result[category].append(Cls)
            return result
        else:
            return l
    
    
    @classmethod
    def menu(cls, action_names, *args, **kwargs):
        """
        Call a certain menu item of the panel
        
        :param list action_names: Example ['File', 'New...']
        :param *args: Positional arguments of the menu call
        :param **kwargs: Keyword arguments of the menu call
        """
        return gui.menu_trigger(cls.category, None, action_names, *args, **kwargs)

    
    @classmethod
    def new(cls, paneltype=None, windowname=None, *args, **kwargs):
        return GuiProxyBase._new(cls.category, paneltype, windowname, *args, **kwargs)
        
        
    @StaticGuiCall        
    def _new(category, paneltype=None, windowname=None, *args, **kwargs):
        return gui.qapp.panels.new(category, paneltype, windowname, *args, **kwargs)

    
    @classmethod
    def selected(cls):
        """
        Return the panel id of the selected panel if this category.
        """
        category = cls.category
        return GuiProxyBase._selected_panid_of_category(category)
        
        
    @StaticGuiCall
    def _selected_panid_of_category(category):
        return gui.qapp.panels.selected(category, panel=False)
        
        
    @classmethod
    def close(Cls, panid=-1):
        """
        Close the selected panel of the category
        """   
        return Cls._close_panel_of_category(panid, Cls.category)
        
        
    @StaticGuiCall
    def _close_panel_of_category(panid, category):        
        panel = gui.qapp.panels.selected(category, panid)
        if not panel is None:
            panel.close_panel()
            return True
        else:
            return False        

#Note that a embeded function is also be static if closure is None            
STATIC = 1   

#Embeded functions with closure not None
ENCLOSED = 2

#Method on object
METHOD = 4

                               
class GuiProxy(object):    
    """
    Communication channel to the gui
    Master and Slave
    Across Procecess and Threads
    """
    
    push_list = []
    
    def __init__(self, qapp=None, master_call_queue=None, master_return_queue=None, process_ready_call=None):
        """
        In case of multiprocessing, there is a gui instance at the master process
        for every child process and an instance at the child process
        
        :param qapp: None in case of child process
        :type qapp: QApplication or None
        :param bool process: True in case of multi processing
        :param multiprocessing.queue master_queue: The event queue from the master process
        """
        self.hooks = dict()
        self.reg_obj = dict()
        self.proxies = dict()
        
        self.block = True
        self._qapp = qapp
                
        self.call_queue = master_call_queue
        self.return_queue = master_return_queue     

        if not process_ready_call is None:
             self.set_func_hook(-2, process_ready_call) 
        
        if not qapp is None:        
            if not self.call_queue is None:
                self.pass_thread = threading.Thread(target=self._pass_to_eventloop, name=f'Handover-{id(self)}', daemon=True)
                self.pass_thread.start()
                
        self.redbull = RedBull()
        self.refresh_proxies()

    @property
    def qapp(self):
        """
        The QApplication instance. Is only visible in the gui thread.
        """
        #Hide it also for threads in the gui app which are not the main
        #User should not use it directly, only by gui_call
        if self.is_main():
            return self._qapp
        else:
            return None
        
    def refresh_proxies(self):
        """
        Add all currently imported Gui Proxies.
        
        :meta private:
        """
        catclasses = GuiProxyBase.derivedClasses()
        for category, classes in catclasses.items():
            proxy = classes[0]()
            name = proxy.attach(self)
            if not name is None:
                self.proxies[name] = proxy
            
    def set_func_hook(self, key, func):
        """
        :meta private:
        """
        self.hooks[key] = func    

    def register_object(self, obj):
        #This is supposed to happen only in the gui thread
        id_obj = id(obj)
        self.reg_obj[id_obj] = [obj] + self.reg_obj.get(id_obj, [])
        return id_obj
        
    def retrieve_object(self, key):        
        lst = self.reg_obj[key]
        func_id = lst.pop()
        if len(lst) == 0:
            self.reg_obj.pop(key)
        return func_id

    def is_main(self):
        return (not self._qapp is None and threading.currentThread().name == 'MainThread')

    def encode_func(self, func, register=False):
        if isinstance(func, (int, tuple)):
            #It is already encoded
            func_id = func
            
        elif isinstance(func, (types.FunctionType, type)):                 
            module = func.__module__
            qualname = func.__qualname__        
            
            if func.__closure__ is None:
                func_id = STATIC, module, qualname
            else:
                closere_id = self.register_object(func.__closure__)
                func_id = ENCLOSED, module, qualname, closere_id            
            
        elif isinstance(func, (types.MethodType, type)):                 
            module = func.__module__
            qualname = func.__qualname__
            self_key = self.register_object(func.__self__)
            func_id = METHOD, module, qualname, self_key
            
        else:
            raise AttributeError(f'Type of {func} is not valid')
            func_id = func.__name__            
                   
        return func_id
            
    def decode_func(self, func_id):        
        if isinstance(func_id, int):
            #For some limit cases, this a a quicker decoding
            #Used for stdout flushing and process_ready
            func = self.hooks[func_id] 
                
        elif isinstance(func_id, tuple):
            lib = importlib.import_module(func_id[1])            
            attrs = func_id[2]
            parts = attrs.split('.')
                
            if func_id[0] in [STATIC, ENCLOSED]:
                tmp = lib
                closure = None if func_id[0] == STATIC else self.retrieve_object(func_id[3])
                for i, attr in enumerate(parts):
                    if attr == '<locals>':
                        tmp = find_nested_func(tmp, parts[i+1], lib.__dict__, closure)
                        break
                    else:
                        tmp = getattr(tmp, attr)
                
                func = tmp                             
                
            elif func_id[0] == METHOD:
                obj = self.retrieve_object(func_id[3])
                tmp = lib                
                for i, attr in enumerate(parts):
                    tmp = getattr(tmp, attr)                    
                func = types.MethodType(tmp, obj)                                   
                
        else:
            #already decoded, was not encoded
            func = func_id                             
                
        return func    

    def gui_call(self, func, *args, **kwargs):
        if self.is_main():
            func = self.decode_func(func)
            return func(*args, **kwargs)
            #return self._qapp.handover.send(self.block, func, *args, **kwargs)
            
        return self._call(func, *args, **kwargs)        
            
    def _call(self, func, *args, **kwargs):                     
        return self._call_base(True, func, *args, **kwargs)            

    def _call_no_wait(self, func, *args, **kwargs):
        return self._call_base(False, func, *args, **kwargs)
            
    def _call_base(self, wait, func, *args, **kwargs):                    
        if self.call_queue is None:
            #Multi Threading Child
            #Direct handover to eventloop
            func = self.decode_func(func)
            return self._qapp.handover.send(self.block, func, *args, **kwargs)
            
        else:
            #Multi Processing Child
            #Handover to Gui Process                
            func = self.encode_func(func)
                
            if wait:
                self.call_queue.put((True, func, args, kwargs))
                value = self.return_queue.get()
                return value
                
            else:
                self.call_queue.put((False, func, args, kwargs))
                return None
        
    def _pass_to_eventloop(self):
        """
        Receive func, args, kwargs from call_queue
        and handover to event loop.
        
        :meta private:
        """
        while True:
            backval, func, args, kwargs = self.call_queue.get()
            func = self.decode_func(func)
            value = self._qapp.handover.send(True, func, *args, **kwargs)
            if backval:
                self.return_queue.put(value)                     
    
    @StaticGuiCall    
    def get_panel_ids(category):
        """
        Returns all current panal ids of a category.
        The last one is the selected one.
        
        :returns: current panal ids 
        :rtype: list
        """
        if category in gui.qapp.panels.keys():
            return list(gui.qapp.panels[category].keys())
        else:
            return []


    @StaticGuiCall    
    def load_layout(name='console'):
        gui.qapp.panels.restore_state_from_config(name)


    def show(self, *args, **argv):
        """
        Shows a object in an suitable panel

        You can give more then one object to display, multiple viewers will be created.

        :param int select: selects which viewer shows the object (start with 1! 0 in GH1)
            negative numbers stands for last selected, or the selected before that, or before that                 
        """        
        objcount = len(args)
        panids = argv.get('select', range(-objcount, 0))
        panid = None
        
        for obj, panid in zip(args, panids):
            if isinstance(obj, np.ndarray):   
                self.img.select(panid)              
                panid = self.img.show(obj)                                                                                 
                
        return panid

    
    @property
    def vs(self):
        """
        The shared array on the current image viewer
        Note that the data is on shared memory.
        ``vs[:]`` or ``vs.ndarray`` to view it as a real numpy array
        """
        try:
            return self.img.vs
        except:
            return None
             
    @property  
    def vr(self):
        """
        The roi sliced numpy array of the current image viewer
        """
        try:
            return self.img.vr
        except:
            return None

    @StaticGuiCall 
    def menu_trigger(category, pandid, action_names, *args, **kwargs):
        """
        Trigger a menu action of a panel.
        
        :param str category: Example 'image'
        :param int id: Example 1
        :param list action_names: Example ['File', 'New...']
        """
        try:
            action = gui.qapp.panels.get_menu_action(category, pandid, action_names)                   
        except KeyError:    
            logger.error(f'Menu action {action_names} not found')
            return
        if len(args) == len(kwargs) == 0:
            action.setData(None)
        else:
            action.setData({'args':args, 'kwargs': kwargs})
        retval = action.trigger()
        action.setData(None)
        return retval

    @StaticGuiCall 
    def history(count=20):   
        """
        The command history
        
        :param int count: Show the last count elements   
        
        :returns: Command history as list
        :rtype: list
        """
        return gui.qapp.history.tail(count)     

    @StaticGuiCall 
    def push(obj):
        """
        Push an object on the gui data stack
        
        :param object obj: A pickable object
        """
        GuiProxy.push_list.append(obj)
        
    @StaticGuiCall       
    def pull():
        """
        Pull and object from the gui data stack
        
        :returns: The object  
        """
        return GuiProxy.push_list.pop(0)        
        
    @StaticGuiCall              
    def exit():
        """
        Exit Gamma Hawk
        """
        gui.qapp.quit()        
        

class GuiMap(object):
    gui_proxies = dict()
    redirects = dict()
    
    def __init__(self):
        pass
        
    @property
    def _gui_proxy(self):
        ident = threading.get_ident()
        ident = GuiMap.redirects.get(ident, ident)        
        return GuiMap.gui_proxies.get(ident, None)
        
    def valid(self):
        return not self._gui_proxy is None
        
    def __dir__(self):
        return self._gui_proxy.__dir__()        
        
    def __repr__(self):
        return self._gui_proxy.__repr__()
        
    def __str__(self):        
        return self._gui_proxy.__str__()        
    
    def __getattr__(self, attr):
        return getattr(self._gui_proxy, attr)
        
    def __setattr__(self, attr, value):
        return setattr(self._gui_proxy, attr, value)

gui = GuiMap()

def register_objects_in_ghawk2_init():
    import gdesk
    gdesk.GuiMap = GuiMap
    gdesk.gui = gui
    gdesk.StaticGuiCall = StaticGuiCall


from ...core.gui_proxy import GuiProxyBase, StaticGuiCall, gui
      
class ScriptWizardProxy(GuiProxyBase):    
    category = 'scriptwiz'
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.script = self
       
    @StaticGuiCall
    def open(templateFile):
        """
        
        """        
        panel = gui.qapp.panels.select_or_new('scriptwiz')                
        panel.openTemplate(templateFile)


# My alternative for multiprocessing Queues
# The original multiprocessing.Queues with locks can only be
# initialized to a process by inheritance
# The shared locks are the problem
# Because there are no locks, to be save
# only a single writer and reader can use this queue

import os
import threading
import collections
import multiprocessing
import _multiprocessing

from multiprocessing.connection import PipeConnection
from multiprocessing.reduction import duplicate, steal_handle
    

class NonDuplexQueue(object):

    def __init__(self):
        self.reader, self.writer = multiprocessing.Pipe(duplex=False)

        self._buffer = collections.deque()
        self._thread = None

    def steal_pipe(self, pid, read_handle, write_handle):
        read_handle = steal_handle(pid, read_handle)
        write_handle = steal_handle(pid, write_handle)

        reader = PipeConnection(read_handle, writable=False)
        writer = PipeConnection(write_handle, readable=False)
        return reader, writer

    def duplicate_pipe(self):
        read_handle = duplicate(self.reader._handle)
        write_handle = duplicate(self.writer._handle)
        return read_handle, write_handle

    def __getstate__(self):
        state = dict()
        state['pid'] = os.getpid()
        read_handle, write_handle = self.duplicate_pipe()
        state['read_handle'] = read_handle
        state['write_handle'] = write_handle
        return state

    def __setstate__(self, state):
        self.reader, self.writer = self.steal_pipe(state['pid'],
            state['read_handle'], state['write_handle'])
        self._buffer = collections.deque()
        self._thread = None

    def empty(self):
        return not self.reader.poll()

    def put(self, data):
        # The _notempty lock can not be pickled and shared
        # between processes (only by inheritance)!
        # So this function should be called after the this queue
        # has been setup between the 2 processes

        if self._thread is None:
            self._notempty = threading.Condition(threading.Lock())
            self._thread = threading.Thread(target=self._feed,
                name='NonDuplexQueueFeed' ,daemon=True)
            self._thread.start()

        with self._notempty:
            self._buffer.append(data)
            self._notempty.notify()

    def get(self, timeout=None):
        if self.reader.poll(timeout):
            return self.reader.recv()
        else:
            raise RuntimeError('Queue is empty')

    def _feed(self):
        while True:
            with self._notempty:
                self._notempty.wait()

            try:
                while True:
                    data = self._buffer.popleft()
                    if data is sentinel:
                        return
                    self.writer.send(data)

            except IndexError:
                pass

    def close(self):
        self.put(sentinel)


import textwrap
import shutil
from collections import OrderedDict

import numpy as np

from .styles import styles

INCR = 100
MAXWIDTH = shutil.get_terminal_size().columns - 4
DEFAULT_STYLE = 'rst-simple'


class ColumnAttributes(object):
    def __init__(self, rectable, colname):
        self._rectable = rectable
        self._colname = colname
        self._attrs = dict()

    def keys(self):
        return tuple(self._attrs.keys())

    def __getitem__(self, key):
        return self._attrs[key]

    def __setitem__(self, key, value):
        self._attrs[key] = value


class ColumnInfo(object):
    def __init__(self, rectable, colname):
        self.rectable = rectable
        self.colname = colname
        self.attrs = ColumnAttributes(self, self.colname)

    def distinct(self):
        return np.unique(self.rectable.sa[self.colname])

    def get_vector(self):
        return self.rectable.sa[self.colname]

    def set_vector(self, value):
        self.rectable.sa[self.colname] = value

    vector = property(get_vector, set_vector)

class More(object):
    def __init__(self, text):
        self.text = text
        self.all_lines = self.text.splitlines()
        self.skip = 0
        self.limit = 20

    @property
    def more(self):
        self.skip += self.limit
        lines = self.all_lines[self.skip:self.skip+self.limit]
        print('\n'.join(lines))

    def __repr__(self):
        lines = self.all_lines[self.skip:self.skip+self.limit]
        return '\n'.join(lines)

class RecordTable(object):
    def __init__(self, fields=[], data=None, size=0, dtype=None):
        """
        fields = ['foo', 'bar', baz']
        or
        fields=[('foo', 'i4'),('bar', 'f4'), ('baz', 'S10')]
        """
        dtype = [(field, 'O') if isinstance(field, str) else field for field in fields] if dtype is None else dtype

        if data is None:
            self._strarr = np.empty(size, dtype=dtype)
        elif isinstance(data, np.ndarray):
            self._strarr = np.array(data, dtype=dtype)
            size = data.size
        else:
            #if data is dataframe
            df = data
            #Remove index??
            self._strarr = df.to_records(index=False)
            size = len(df)

        self._stop = size
        self.base = None
        self.active_row = None
        self.maxcolwidth = None
        self.maxwidth = MAXWIDTH
        self.keepnextlines = True

        self._colinfo = OrderedDict()

        for colname in self._strarr.dtype.names:
            self._colinfo[colname] = ColumnInfo(self, colname)

        self._index_colnames = []

    def set_index(self, colnames):
        if isinstance(colnames, str):
            self._index_colnames = [colnames]
        else:
            assert isinstance(colnames, tuple) or isinstance(colnames, list)
            self._index_colnames = list(colnames)

    def get_column_info(self, colname):
        return self._colinfo[colname]

    @classmethod
    def empty(cls, dtype, size=0):
        self = cls()

    def get_size(self):
        return self._stop

    def set_size(self, length):
        incr = (length - len(self._strarr))
        min_incr = max(incr, INCR)
        self._stop = length
        if self._stop > len(self._strarr):
            self._strarr = np.resize(self._strarr, len(self._strarr) + min_incr)

    size = property(get_size, set_size)

    @property
    def sa(self):
        return self._strarr[0:self._stop]

    @property
    def ra(self):
        return np.rec.array(self.sa, copy=False)

    def copy(self):
        return RecordTable(data=self.sa, dtype=self.sa.dtype)

    @property
    def df(self):
        from pandas import DataFrame
        return DataFrame(self.sa)

    def set_header(self, header, dtypes=None):
        curr_colnames = self.colnames

        if len(header) > len(curr_colnames):
            for colname in header[len(curr_colnames):]:
                self.add_column(colname)

        self.colnames = header

    def __truediv__(self, colname):
        self.add_column(colname)
        return self

    def add_record(self, *record_items, **record_dict):
        if len(record_dict) > 0:
            self.add_record_dict(record_dict)
        else:
            self.add_record_tuple(record_items)

    def add_record_dict(self, record_as_dict):
        record = tuple(record_as_dict[field] for field in self.colnames)
        self.add_record_tuple(record)

    def add_record_tuple(self, record):
        self[self.size] = record

    def add_row(self, row):
        self.add_record_tuple(tuple(row))

    def add_column(self, name, title=None, data=None, dtype=object, method='linear'):
        source = self.sa
        existing_descr = source.dtype.descr
        existing_descr = [] if existing_descr == [('','<f8')] else existing_descr
        self._strarr = np.empty(len(source), dtype=existing_descr+[((title,name), dtype)])
        colnames = source.dtype.names
        if not colnames is None:
            for colname in colnames:
                self._strarr[colname] = source[colname]
        if not data is None:
            m = len(data)
            if method == 'linear':
                if m > self.size:
                    self.size = m
                    self.sa[name] = data
                elif m < self.size:
                    self.sa[name][:m] = data
                else:
                    self.sa[name] = data
            elif method == 'orthogonal':
                prior_n = self.size
                self.size = self.size * m
                self._strarr = np.tile(self.sa, m)
                self.sa[name] = np.repeat(data, prior_n)

    def __getitem__(self, slices):
        if isinstance(slices, RecordTable):
            slices = slices.sa.view(bool)

        elif isinstance(slices, str):
            #result = self.sa.__getitem__(slices)
            result = self.get_column_info(slices)
            return result

        result = self.sa.__getitem__(slices)

        if isinstance(result, np.ndarray):
            arr = RecordTable(dtype=result.dtype)
            arr._strarr = result
            arr._stop = result.size
            return arr

        else:
            return result

    def __setitem__(self, slices, values):
        if isinstance(slices, int) and slices > (self.size-1):
            self.size = slices + 1

        self.sa.__setitem__(slices, values)

    def get_colnames(self):
        if self._strarr.dtype.names is None:
            return []
        return self._strarr.dtype.names

    def set_colnames(self, colnames):
        self._strarr.dtype.names = colnames

    colnames = property(get_colnames, set_colnames)
    field_names = colnames

    def __str__(self):
        return self.tabulate(style=DEFAULT_STYLE)

    def _operate_as_sa(self, func, other):
        if isinstance(other, RecordTable):
            other = other.sa
        result = func(self.sa, other)
        return RecordTable(result, dtype=[('mask', bool)])

    def __or__(self, cell):
        if not self.active_row is None:
            self.active_row.end()
        self.active_row = self.new_record()
        self.active_row.add_cell(cell)
        return self.active_row

    def __eq__(self, other):
        return self._operate_as_sa(np.ndarray.__eq__, other)

    def __ne__(self, other):
        return self._operate_as_sa(np.ndarray.__ne__, other)

    def __lt__(self, other):
        return self._operate_as_sa(np.ndarray.__lt__, other)

    def __le__(self, other):
        return self._operate_as_sa(np.ndarray.__le__, other)

    def __gt__(self, other):
        return self._operate_as_sa(np.ndarray.__gt__, other)

    def __ge__(self, other):
        return self._operate_as_sa(np.ndarray.__ge__, other)

    def __and__(self, other):
        return self._operate_as_sa(np.ndarray.__and__, other)

    def __or__(self, other):
        return self._operate_as_sa(np.ndarray.__or__, other)

    def __xor__(self, other):
        return self._operate_as_sa(np.ndarray.__xor__, other)

    def __len__(self):
        return self.get_size()

    def new_record(self):
        return Record(self)

    def end(self):
        if not self.active_row is None:
            self.active_row.end()
        self.active_row = None

    def tabulate(self, style='rst-simple', haligns='l', valigns='t', maxwidth=None, debug=False, auto_index='pos'):

        if not style in styles.keys():
            raise AttributeError(f"Style {style} doesn't exists. Choose between:\n{styles.keys()}")

        style_param = styles[style]
        art = style_param['art']
        hhaligns = style_param['hhaligns']

        hhaligns = hhaligns.ljust(len(self.colnames), hhaligns[-1])
        haligns = haligns.ljust(len(self.colnames), haligns[-1])
        valigns = valigns.ljust(len(self.colnames), valigns[-1])

        align_map = dict(l = '<', c = '^', r = '>')

        def wrap_paragraph(text):
            lines = []
            for line in text.splitlines():
                lines.extend(txtwr.wrap(line))
            return [''] if len(lines) == 0 else lines

        def wrap_ignore_new_lines(text):
            lines = txtwr.wrap(text)
            return [''] if len(lines) == 0 else lines

        def lines_html(text):
            return [text.replace('\n', '<br>')]

        dtype_obj = [(colname, 'O') for colname in self.colnames]
        if auto_index:
            dtype_obj = [('AUTO_INDEX', 'O')] + dtype_obj
        arr_str = np.empty(self.size, dtype=dtype_obj)

        if auto_index:
            arr_str['AUTO_INDEX'] = [[str(item)] for item in range(self.size)]

        maxwidth = maxwidth or self.maxwidth
        maxcolwidth = maxwidth if self.maxcolwidth is None else self.maxcolwidth

        if auto_index:
            maxcolwidth_index = max(3, max(max(len(line) for line in field) for field in arr_str['AUTO_INDEX']))
            maxcolwidth = min(maxwidth - len(art[3][0]) - len(art[3][3]) - maxcolwidth_index - len(art[3][2]), maxcolwidth)
        else:
            maxcolwidth = min(maxwidth - len(art[3][0]) - len(art[3][3]), maxcolwidth)

        txtwr = textwrap.TextWrapper(maxcolwidth)

        if style == 'html':
            lines_wrapper = lines_html
        else:
            if self.keepnextlines:
                lines_wrapper = wrap_paragraph
            else:
                lines_wrapper = wrap_ignore_new_lines

        for colname in self.colnames:
            arr_str[colname] = [lines_wrapper(str(item)) for item in self.sa[colname]]

        colnames = list(self.colnames)
        titles = list(self.colnames)

        if auto_index:
            colnames = ['AUTO_INDEX'] + colnames
            titles = [auto_index] + titles
            hhaligns = 'r' + hhaligns
            haligns = 'r' + haligns
            valigns = 't' + valigns

        if self.size == 0:
            colwidths = [len(title) for title in titles]
        else:
            colwidths = []
            for colname, title in zip(colnames, titles):
                colwidths.append(max(len(title), max(max(len(line) for line in field) for field in arr_str[colname])))

        result = ''

        def add_line(s):
            nonlocal result
            result += s + '\n'

        def calc_width_for_selection(selection):
            width = len(art[3][0]) + len(art[3][3])
            width += sum(colwidths[pos] + len(art[3][2]) for pos in selection)
            width -= len(art[3][2])
            return width

        selections = []

        if auto_index:
            minumum_selection = [0]
        else:
            minumum_selection = []

        leftover_selection = list(range(len(colnames)))
        [leftover_selection.pop(pos) for pos in minumum_selection]

        start = 0
        stop = 0

        while stop <= len(leftover_selection):
            stop += 1
            scan_selection = minumum_selection.copy()
            scan_selection.extend(leftover_selection[start:stop])
            width = calc_width_for_selection(scan_selection)

            if debug:
                print(f'Start {start}; Stop {stop}; Selection {scan_selection}; Width {width}')

            if (width > maxwidth):
                selections.append(scan_selection[:-1])
                start = stop - 1

            elif stop > len(leftover_selection):
                selections.append(scan_selection)

            if debug:
                print(selections)


        for selection in selections:
            if debug:
                add_line(''.join(str(i % 10) for i in range(maxwidth)))

            selcolwidth = [colwidths[pos] for pos in selection]
            selhaligns = [haligns[pos] for pos in selection]
            selhhaligns = [hhaligns[pos] for pos in selection]
            seltitles = [titles[pos] for pos in selection]

            if not art[0] is None:
                add_line(art[0][0] + art[0][2].join(art[0][1] * colwidth for colwidth in selcolwidth) + art[0][3])

            if not art[1] is None:
                head_template = art[1][0] + art[1][2].join(f"{{{i}:{align_map[align]}{width}}}" for i, (width, align) in enumerate(zip(selcolwidth, selhhaligns))) + art[1][3]
                add_line(head_template.format(*seltitles))

            if not art[2] is None:
                add_line(art[2][0] + art[2][2].join(art[2][1] * colwidth for colwidth in selcolwidth) + art[2][3])

            row_template = art[3][0] + art[3][2].join(f"{{{i}:{align_map[align]}{width}}}" for i, (width, align) in enumerate(zip(selcolwidth, selhaligns))) + art[3][3]

            for ind, record in enumerate(arr_str):
                selrecord = [record[pos] for pos in selection]
                last = ind == (arr_str.size-1)

                line_counts = [len(field) for field in selrecord]
                max_line_count = max(line_counts)

                for linenr in range(max_line_count):
                    line = []
                    for fnr, field in enumerate(selrecord):
                        valign = valigns[fnr]
                        if valign == 't':
                            start = 0
                            stop = len(field)

                        elif valign == 'c':
                            start = (max_line_count - len(field)) // 2
                            stop = start + len(field)

                        elif valign == 'b':
                            start = max_line_count - len(field)
                            stop = max_line_count

                        if start <= linenr < stop:
                            line.append(field[linenr-start])
                        else:
                            line.append('')

                    add_line(row_template.format(*line))

                if not last and not art[4] is None:
                    add_line(art[4][0] + art[4][2].join(art[4][1] * colwidth for colwidth in selcolwidth) + art[4][3])
                elif last and not art[5] is None:
                    add_line(art[5][0] + art[5][2].join(art[5][1] * colwidth for colwidth in selcolwidth) + art[5][3])

        return result

    def get_html_string(self):
        htmlstr = '<table>\n' + self.tabulate(style='html', maxwidth=65536) + '</table>\n'
        return htmlstr

    def to_clipboard_html(self):
        from .htmlclip import PutHtml
        htmlstr = '<table>\n' + self.tabulate(style='html') + '</table>\n'
        PutHtml(htmlstr)

    def to_clipboard_pre_formatted(self):
        import win32clipboard as cb
        try:
            cb.OpenClipboard()
            cb.EmptyClipboard()
            prestr = self.tabulate()
            cb.SetClipboardText(prestr)
        finally:
            cb.CloseClipboard()

    def show(self, ipython=False):
        if ipython:
            ht = self.get_html_string()
            from IPython.core.display import display, HTML
            display(HTML(ht))
        else:
            print(self)


class Record(object):
    def __init__(self, table):
        self.table = table
        self.cells = []

    def add_cell(self, cell):
        self.cells.append(cell)

    def __or__(self, cell):
        self.add_cell(cell)
        return self

    def end(self):
        self.table.add_record(self.cells)






import os
import textwrap
import psutil
import time
import logging
from pathlib import Path

from qtpy import QtCore, QtGui, QtWidgets

from qtpy.QtCore import Qt, QTimer, QSize
from qtpy.QtGui import QFont, QFontMetrics, QTextCursor, QTextOption, QPainter, QTextCharFormat
from qtpy.QtWidgets import (QAction, QMainWindow, QPlainTextEdit, QSplitter, QVBoxLayout, QLineEdit, QLabel,
    QMessageBox, QTextEdit, QWidget, QStyle, QStyleFactory, QApplication, QCompleter, QComboBox)

from ... import config, gui
from ...core import tasks
from ...core.shellmod import Shell
from ...panels.base import BasePanel, selectThisPanel, CheckMenu
from ...dialogs.formlayout import fedit
from ...dialogs.base import messageBox
from ...dialogs.editpaths import EditPaths
from ...utils.syntax_light import analyze_python, ansi_highlight
from ...utils.ansi_code_processor import QtAnsiCodeProcessor
from ...gcore.utils import getMenuAction

respath = Path(config['respath'])
logger = logging.getLogger(__name__) 

MAXCHARPERLINE = 10000  #Only for ansi mode
#PREFIXWIDTH = 30  # width of the left side line numbers

ANSI_ESCAPE_SYNTAX_HIGHLIGHT = {
    'comment': ('\033[38;5;2m', '\033[39m'),
    'string': ('\033[38;5;8m', '\033[39m'),
    'docstring': ('\033[38;5;208m', '\033[39m'),
    'keyword': ('\033[38;5;12m', '\033[39m'),
    'builtin': ('\033[38;5;6m', '\033[39m'),
    'definition': ('\033[38;5;12m', '\033[39m'),
    'defname': ('\033[38;5;13m', '\033[39m'),
    'operator': ('\033[38;5;4m', '\033[39m'),
}

ESC = '\033['
ERROR_PREFIX = ESC + '38;5;9m'
ERROR_SUFFIX = ESC + '0m'


class LineNumberArea(QWidget):

    def __init__(self, textEditor):
        QWidget.__init__(self, textEditor)
        self.textEditor=textEditor
        self.prefix_color = Qt.lightGray
        self.prefix_font_color = Qt.black
        self.update_font()

        self._firstlinecode = [' >>> ']
        self.blinks = [':',' ']

        self.painter = QPainter()

        self.promptTimer = QTimer(self)
        self.promptTimer.timeout.connect(self.nextPrompt)

        self.profile_start = time.monotonic()
        self.profile_threshold = 10
        self.profile_enabled = False

    def start_profiling(self):
        self.profile_start = time.monotonic()
        self.profile_enabled = True

    def stop_profiling(self):
        self.profile_enabled = False

    def update_font(self):
        self.setFont(self.textEditor.font())
        self.fontmetric = QFontMetrics(self.textEditor.font())
        self.prefixwidth = self.fontmetric.width('12345')
        self.prefixheight = self.fontmetric.height()
        self.setFixedWidth(self.prefixwidth)
        self.textEditor.setViewportMargins(self.prefixwidth, 0, 0, 0)

    def set_firstlinecode(self, prefices):
        self._firstlinecode = prefices

        if len(self._firstlinecode) > 1:
            self.promptTimer.start(500)
        else:
            self.promptTimer.stop()

    def get_firstlinecode(self):
        if self.profile_enabled:
            elapsed = time.monotonic() - self.profile_start
            if elapsed > self.profile_threshold:
                if elapsed >= 3600:
                    profile = time.strftime("%H:%M", time.gmtime(elapsed))
                else:
                    profile = time.strftime("%M:%S", time.gmtime(elapsed))
                return profile[:-3] + self.blinks[0] + profile[-2:]
        return self._firstlinecode[0]

    firstlinecode = property(get_firstlinecode, set_firstlinecode)

    def nextPrompt(self):
        self._firstlinecode = self._firstlinecode[1:] + self._firstlinecode[:1]
        self.blinks = self.blinks[1:] + self.blinks[:1]
        self.repaint()

    def paintEvent(self, event):
        cursor = self.textEditor.cursorForPosition(self.textEditor.viewport().pos())

        painter = self.painter
        try:
            painter.begin(self)
            painter.fillRect(event.rect(), self.prefix_color)

            for i in range(100):
                blockNumber = cursor.block().blockNumber()

                if blockNumber == 0:
                    code = self.firstlinecode
                else:
                    code = str(blockNumber + 1)

                painter.setPen(self.prefix_font_color)
                y = self.textEditor.cursorRect(cursor).y() + self.textEditor.viewport().pos().y() - 2
                painter.drawText(0, y, self.prefixwidth, self.prefixheight, Qt.AlignRight, code)

                if y > event.rect().bottom():
                    break
                if not cursor.block().next().isValid():
                    break

                cursor.movePosition(cursor.NextBlock)
        finally:
            painter.end()

    def sizeHint(self):
        return QSize(self.prefixwidth, 0)

class StdInputPanel(QPlainTextEdit):
    def __init__(self, parent, task, outputPanel):
        super().__init__(parent = parent)
        self.task = task
        self.outputPanel = outputPanel
        self.qapp = QApplication.instance()

        self.prior_cmd_id = None
        self.hist_prefix = None

        self.configure(config)
        self.lineNumberArea=LineNumberArea(self)

        #self.setFocusPolicy(Qt.StrongFocus)

        self.styles = dict()
        self.styles['interprete'] = "background-color:white;"
        self.styles['wait'] = "background-color:#DDBBBB;"
        self.styles['running'] = "background-color:#FFFFE0;"
        self.styles['input'] = "background-color:#BBBBDD;"
        self.styles['ended'] = "background-color:#EFEFEF;"

        self.setMinimumHeight(32)        

        self.mode = 'interprete'
        self.heightHint = 100

    def configure(self, config):
        console_font = QFont('Consolas', pointSize=config['console']['fontsize'])
        self.setFont(console_font)

        if config['console']['wrap']:
            self.setWordWrapMode(QTextOption.WordWrap)
        else:
            self.setWordWrapMode(QTextOption.NoWrap)

        self.setUndoRedoEnabled(True)

    def sizeHint(self):
        return QtCore.QSize(200, self.heightHint)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        rect=self.contentsRect()
        self.lineNumberArea.setGeometry(rect.x(),rect.y(), self.lineNumberArea.prefixwidth, rect.height())

    def keyPressEvent(self, event):
        self.lineNumberArea.update(0, 0, self.lineNumberArea.width(), self.lineNumberArea.height())

        #Is the enter pressed on numeric keypad or on the base keypad
        key_enter = (event.key() == Qt.Key_Return) or \
            (event.key() == Qt.Key_Enter)

        #left or right shift
        modifiers = event.nativeModifiers()
        key_shift = modifiers & 1 == 1 or  modifiers & 16 == 16
        #left or right ctrl
        key_ctrl = modifiers & 2 == 2 or modifiers & 32 == 32

        if not event.key() in [Qt.Key_Up, Qt.Key_Down]:
            self.prior_cmd_id = None
            self.hist_prefix = None

        if key_enter:
            if key_ctrl:
                self.execute_commands()
            elif key_shift:
                self.textCursor().insertBlock()
            else:
                if self.blockCount() == 1:
                    self.execute_commands()
                else:
                    if self.lastLineIsEmpty():
                        self.execute_commands()
                    else:
                        self.textCursor().insertBlock()

        elif event.key() == Qt.Key_Tab:
            if key_ctrl or key_shift:
                self.startAutoCompleter(wild=True)
            else:
                self.startAutoCompleter()

        elif event.key() == Qt.Key_Up and self.textCursor().block().blockNumber() == 0:
            if self.hist_prefix is None:
                self.hist_prefix = self.toPlainText()
            self.prior_cmd_id, cmd = self.qapp.history.retrievecmd(self.hist_prefix, self.prior_cmd_id, distinct=True, back=True, prefix=not key_ctrl)

            self.setPlainText(cmd)
            self.moveCursorToEndOfBlock()

        elif event.key() == Qt.Key_Down and self.textCursor().block().blockNumber() == (self.blockCount()-1):
            if self.hist_prefix is None:
                self.hist_prefix = self.toPlainText()

            self.prior_cmd_id, cmd = self.qapp.history.retrievecmd(self.hist_prefix, self.prior_cmd_id, distinct=True, back=False, prefix=not key_ctrl)

            self.setPlainText(cmd)
            self.moveCursorToEndOfDoc()

        else:
            super().keyPressEvent(event)

    def lastLineIsEmpty(self):
        self.cursor=self.textCursor()
        if self.cursor.block().blockNumber() != (self.blockCount() - 1):
            return False

        self.cursor.movePosition(self.cursor.EndOfLine, self.cursor.KeepAnchor)
        curdocpos = self.cursor.position()
        self.cursor.movePosition(self.cursor.StartOfLine, self.cursor.KeepAnchor)
        startlinepos = self.cursor.position()

        return curdocpos == startlinepos

    def startAutoCompleter(self, wild=False):

        #delims = ' \t\n\\"\'`@$><=;|&{('
        delims = ' \t\n\\`@$><=;|&{('

        current_text = self.toPlainText()

        try:
            pos = min(current_text[::-1].index(c) for c in delims if c in current_text)
            part = current_text[-pos:]
        except:
            pos = len(current_text)

        if pos == 0:
            self.insertText('    ')
            return

        self.keep, self.part = current_text[:-pos], current_text[-pos:]

        self.outputPanel.addText(f'{self.part}*\n')

        max_items = config['console']['max_complete']
        self.task.call_func(Shell.get_completer_data, (self.part, max_items, wild), self.response_to_autocomplete)

    def moveCursorToEndOfBlock(self):
        cursor=self.textCursor()
        cursor.movePosition(cursor.EndOfBlock)
        self.setTextCursor(cursor)

    def moveCursorToEndOfDoc(self):
        cursor=self.textCursor()
        cursor.movePosition(cursor.End)
        self.setTextCursor(cursor)

    def execute_commands(self, cmd=None):
        if cmd is None:
            cmd = self.toPlainText()

        if self.mode in ['interprete', 'running']:
            cmd = textwrap.dedent(cmd)
            histcmd = cmd

            if cmd.startswith('%'):
                cmd = 'shell.magic(r"""' + cmd[1:]+ '""")'

            elif cmd.startswith('!!'):
                cmd = 'shell.popen(r"""' + cmd[2:] + '""", shell=False)'

            elif cmd.startswith('!'):
                cmd = 'shell.popen(r"""' + cmd[1:] + '""", shell=True)'

            elif cmd.endswith('??'):
                cmd = 'shell.edit(' + cmd[:-2] + ')'

            elif cmd.endswith('?'):
                histcmd = cmd
                cmd = 'help(' + cmd[:-1] + ')'

            elif cmd.endswith('!!'):
                cmd = 'shell.pprint(' + cmd[:-2] + ')'

            elif cmd.endswith('!'):
                cmd = 'print(' + cmd[:-1] + ')'

            if cmd.count('\n') == 0:
                prefix = '\033[48;5;7m>>>\033[0m \033[1m'
                suffix = '\033[0m\n'
            else:
                prefix = '\033[48;5;7m>>>\n\033[0m\033[1m'
                suffix = '\033[0m\n\033[48;5;7m<<<\033[0m\n'

            try:
                cmdecho = ansi_highlight(analyze_python(cmd), colors=ANSI_ESCAPE_SYNTAX_HIGHLIGHT)
            except:
                cmdecho = cmd

            self.outputPanel.addAnsiText(prefix + cmdecho + suffix)

            self.qapp.history.logcmd(histcmd)
            self.task.send_command(cmd, self.retval_ready)

            #self.set_mode('running')

            self.clear()

        elif self.mode == 'input':

            if cmd.count('\n') == 0:
                prefix = '\033[48;5;7m>?\033[0m \033[1m'
                suffix = '\033[0m\n'
            else:
                prefix = '\033[48;5;7m>?\n\033[0m\033[1m'
                suffix = '\033[0m\n\033[48;5;7m?<\033[0m\n'

            self.outputPanel.addAnsiText(prefix + cmd + suffix)
            self.task.send_input(cmd)
            self.clear()

    def retval_ready(self,  mode, error_code, result):
        if mode == 'interprete':
            if error_code == 1:
                self.outputPanel.addAnsiText(f'\033[38;5;9msyntax error: {result}\033[0m\n')

            elif error_code == 2:
                self.outputPanel.addAnsiText(f'\033[38;5;9mincomplete error: {result}\033[0m\n')
                self.setPlainText(result + '\n')
                self.moveCursorToEndOfDoc()

        self.set_mode('interprete')

    def set_mode(self, mode='interprete'):
        if mode == 'wait':
            self.setStyleSheet(self.styles[mode])
            self.mode = mode
            self.lineNumberArea.firstlinecode = [' ... ',
                                                 '  ...',
                                                 '   ..',
                                                 '    .',
                                                 '     ',
                                                 '.    ',
                                                 '..   ',
                                                 '...  ']
            self.lineNumberArea.start_profiling()
            self.setReadOnly(True)

        if mode == 'running':
            self.setStyleSheet(self.styles[mode])
            self.mode = mode
            self.lineNumberArea.firstlinecode = [' ... ',
                                                 '  ...',
                                                 '   ..',
                                                 '    .',
                                                 '     ',
                                                 '.    ',
                                                 '..   ',
                                                 '...  ']
            self.lineNumberArea.start_profiling()
            self.setReadOnly(False)

        elif mode == 'interprete':
            self.setStyleSheet(self.styles[mode])
            self.mode = mode
            self.lineNumberArea.firstlinecode = [' >>> ']
            self.lineNumberArea.stop_profiling()
            self.setReadOnly(False)

        elif mode == 'input':
            self.setStyleSheet(self.styles[mode])
            self.mode = mode
            self.lineNumberArea.firstlinecode = ['>?   ', '   >?']
            self.lineNumberArea.stop_profiling()
            self.setReadOnly(False)

        elif mode == 'ended':
            self.outputPanel.setStyleSheet(self.styles[mode])
            self.hide()
            self.mode = mode
            self.lineNumberArea.firstlinecode = ['Ended']
            self.lineNumberArea.stop_profiling()
            self.setReadOnly(True)

        self.lineNumberArea.update()

    def response_to_autocomplete(self, tag, error_code, items):
        if len(items) == 0:
            return

        if len(items) == 1:
            self.setPlainText(self.keep + items[0])
            self.moveCursorToEndOfDoc()

        else:
            for item in items:
               self.outputPanel.addText(f'{item}\n')

            commonprefix = os.path.commonprefix(items)

            if commonprefix != self.part:
                self.setPlainText(self.keep + commonprefix)
                self.moveCursorToEndOfDoc()

    def wheelEvent(self, event):
        modifiers = event.modifiers()
        #key_ctrl = modifiers & 2 == 2 or modifiers & 32 == 32
        key_ctrl = modifiers & Qt.ControlModifier
        #key_ctrl = True

        #print(modifiers)

        if key_ctrl and event.delta() < 0:
            font = self.font()
            font.setPointSize(font.pointSize()-1)
            self.setFont(font)

        elif key_ctrl and event.delta() > 0:
            font = self.font()
            font.setPointSize(font.pointSize()+1)
            self.setFont(font)

        self.lineNumberArea.update_font()

        if not key_ctrl:
            super().wheelEvent(event)

    # def focusInEvent(self, event):
        # selectThisPanel(self)
        # super().focusInEvent(event)

    def replaceSelected(self, text):
        cursor = self.textCursor()
        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        if cursor.hasSelection():
            cursor.insertText(text)
            #end = cursor.position()
            end = start + len(text)

        #QTextCursor.KeepAnchor
        cursor.setPosition(start)
        cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.setTextCursor(cursor)

    def insertText(self, text):
        cursor=self.textCursor()
        cursor.insertText(text)

    def addText(self, text):
        cursor=self.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.moveCursor(QTextCursor.End)


class StdPlainOutputPanel(QPlainTextEdit):
    def __init__(self, parent, stdout_queue):
        super().__init__(parent = parent)
        self.stdout_queue = stdout_queue
        self.setReadOnly(True)
        self._ansi_processor = None
        self.auto_scroll = True
        self.configure(config)

    @property
    def panel(self):
        return self.parent().parent().parent()

    def configure(self, config):
        console_font = QFont('Consolas', pointSize=config['console']['fontsize'])
        self.setFont(console_font)

        if config['console']['wrap']:
            self.setWordWrapMode(QTextOption.WordWrap)
        else:
            self.setWordWrapMode(QTextOption.NoWrap)

        self.setMaximumBlockCount(config['console']['maxblockcount'])

    def flush(self):
        text = ''
        while not self.stdout_queue.empty():
            data = self.stdout_queue.get()
            if isinstance(data, tuple):
                ttype, content = data
            else:
                ttype = 'raw'
                content = data
            text += content
        if not text == '':
            if ttype == 'raw':
                self.addText(text)
            elif ttype == 'error':
                self.addAnsiText(f'{ERROR_PREFIX}{text}{ERROR_SUFFIX}')
                self.panel.show_me()
            elif ttype == 'ansi':
                self.addAnsiText(text)
                #Some bug, reset format to normal
                #self.addAnsiText('\n')

    def addText(self, text):
        cursor=self.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        if self.auto_scroll:
            self.moveCursor(QTextCursor.End)

    def addAnsiText(self, text):
        cursor = self.textCursor()
        cursor.movePosition(QTextCursor.End)
        self._insert_ansi_escape_text(cursor, text)
        if self.auto_scroll:
            self.moveCursor(QTextCursor.End)

    def _insert_ansi_escape_text(self, cursor, text):
        cursor.beginEditBlock()

        if self._ansi_processor == None:
            self._ansi_processor = QtAnsiCodeProcessor()

        if len(text) > 0:
            if text[0] == '\r':
                # BUG when no character is preceding the <CR>
                # editor goes to some string state
                text = ' ' + text

        for substring in self._ansi_processor.split_string(text):
            for act in self._ansi_processor.actions:

                # Unlike real terminal emulators, we don't distinguish
                # between the screen and the scrollback buffer. A screen
                # erase request clears everything.
                if act.action == 'erase' and act.area == 'screen':
                    cursor.select(QTextCursor.Document)
                    cursor.removeSelectedText()

                # Simulate a form feed by scrolling just past the last line.
                elif act.action == 'scroll' and act.unit == 'page':
                    cursor.insertText('\n')
                    cursor.endEditBlock()
                    self._set_top_cursor(cursor)
                    cursor.joinPreviousEditBlock()
                    cursor.deletePreviousChar()

                elif act.action == 'carriage-return':
                    cursor.movePosition(
                        cursor.StartOfLine, cursor.KeepAnchor)

                elif act.action == 'beep':
                    gui.qapp.beep()

                elif act.action == 'backspace':
                    cursor.movePosition(
                        cursor.PreviousCharacter, cursor.KeepAnchor)

                elif act.action == 'newline':
                    cursor.movePosition(cursor.EndOfLine)

            format = self._ansi_processor.get_format()

            # This doesn't seem to work with special characters
            # backspace seems to disable the output, no recovery

            selection = cursor.selectedText()
            if len(selection) == 0:
                if substring is None:
                    pass
                elif len(substring) > MAXCHARPERLINE:
                    cursor.insertText(substring[:MAXCHARPERLINE] + "...%d chars not displayed" % (len(substring) - MAXCHARPERLINE), format)
                else:
                    cursor.insertText(substring, format)

            elif substring is not None:
                # BS and CR are treated as a change in print
                # position, rather than a backwards character
                # deletion for output equivalence with (I)Python
                # terminal.
                if len(substring) >= len(selection):
                    cursor.insertText(substring, format)
                else:
                    old_text = selection[len(substring):]
                    cursor.insertText(substring + old_text, format)
                    cursor.movePosition(cursor.PreviousCharacter,
                           cursor.KeepAnchor, len(old_text))

        cursor.setBlockCharFormat(QTextCharFormat())
        cursor.endEditBlock()


class StdioFrame(QWidget):
    """
    A Window with standard input and output panels
    For threads and processes.
    """
    def __init__(self, parent, title, task):
        super().__init__(parent=parent)

        self.task = task

        self.setWindowTitle(title)
        self.stdOutputPanel = StdPlainOutputPanel(self, task.stdout_queue)
        self.stdInputPanel = StdInputPanel(self, task, self.stdOutputPanel)

        task.set_flusher(self.stdOutputPanel.flush)

        splitter = QSplitter(Qt.Vertical, self)
        splitter.addWidget(self.stdOutputPanel)
        splitter.addWidget(self.stdInputPanel)

        vbox = QVBoxLayout()
        self.setLayout(vbox)
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(splitter)

    def keyPressEvent(self, event):
        pass

class WrapCaller(object):
    def __init__(self, caller, *args, **kwargs):
        self.caller = caller
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.caller(*self.args, **self.kwargs)

class RecentMenu(QtWidgets.QMenu):
    def __init__(self, parent=None):
        super().__init__('Recent', parent)
        self.panel = self.parent()

    def showEvent(self, event):
        self.initactions()

    def initactions(self):
        self.clear()
        self.actions = []

        for rowid, timestamp, path in gui.qapp.history.yield_recent_paths(category='console'):
            action = QtWidgets.QAction(path, self)
            action.triggered.connect(WrapCaller(self.panel.openFile, path))
            self.addAction(action)
            self.actions.append(action)

class Console(BasePanel):
    panelCategory = 'console'
    userVisible = False

    classIconFile = str(respath / 'icons' / 'px16' / 'application_xp_terminal.png')

    def __init__(self, parent, panid, task):
        super().__init__(parent, panid, 'console')

        self.stdio = StdioFrame(self, '', task)
        task.console = self
        self.setCentralWidget(self.stdio)
        self.createMenus()
        self.createStatusBar()

        self.stdio.stdOutputPanel.flush()
        self.setFocusPolicy(Qt.StrongFocus)

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.updateProcessInfo)
        self.timer.start(config['system info period'])

    @property
    def task(self):
        return self.stdio.task

    def createMenus(self):
        self.fileMenu = self.menuBar().addMenu("&File")
        #self.executionMenu = self.menuBar().addMenu("&Execution")

        self.executionMenu = CheckMenu("&Execution", self.menuBar())
        self.menuBar().addMenu(self.executionMenu)

        self.addMenuItem(self.fileMenu, "Open File", self.openFileDialog,
            statusTip="Open a file", icon = 'folder_vertical_document.png')

        self.addMenuItem(self.fileMenu, 'New Thread', self.newThread)
        self.fileMenu.addMenu(RecentMenu(self))
        self.addMenuItem(self.fileMenu, "Close", self.close_panel,
            statusTip = "Close this Thread",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'cross.png')))

        traceMenu = QtWidgets.QMenu('tracing')
        traceMenu.addAction(QAction("Enable Tracing", self, triggered=lambda: self.task.set_tracing(True),
            statusTip="Enable Tracing, needed for sync breaks"))
        traceMenu.addAction(QAction("Disable Tracing", self, triggered=lambda: self.task.set_tracing(False),
            statusTip="Disable Tracing"))
        self.addMenuItem(traceMenu, "Enable Timeit", lambda: self.task.set_timeit(True),
            statusTip="Show elapsed time report at end of execution",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'time_red.png')))
        traceMenu.addAction(QAction("Disable Timeit", self, triggered=lambda: self.task.set_timeit(False),
            statusTip="Disable Timeit"))
        traceMenu.addAction(QAction("Enable profiling", self, triggered=lambda: self.task.enable_profiling(),
            statusTip="One time profiling of the next command"))
        self.executionMenu.addMenu(traceMenu)

        self.addMenuItem(self.executionMenu, 'Check Flow Alive', self.checkAlive,
            statusTip="Check if the flow loop is still alive")
        self.addMenuItem(self.executionMenu, 'Print Trace', self.stdio.task.print_trace,
            statusTip="Print the trace of the current execution frame")
        self.addMenuItem(self.executionMenu, 'Print Locals', self.stdio.task.print_locals,
            statusTip="Print the locals of the current namespace")
        self.addMenuItem(self.executionMenu, 'Sync Break', self.syncBreak,
            statusTip="Send a synchronous break, tracing should be active",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'cog_stop.png')))
        self.addMenuItem(self.executionMenu, 'Async Break', self.asyncBreak,
            statusTip="Send a asynchronous break",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'cog_stop.png')))
        self.addMenuItem(self.executionMenu, 'System Exit Thread', self.stdio.task.system_exit,
            statusTip="System Exit")

        self.executionMenu.addSeparator()

        # self.addMenuItem(self.executionMenu, 'Pause Process', self.suspendResumeProcess,
            # statusTip="Suspend this windows process")
        self.addMenuItem(self.executionMenu, 'Kill Process', self.killProcess,
            statusTip="Kill this Python Process and its threads",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'scull.png')))

        self.viewMenu = CheckMenu("&View", self.menuBar())
        self.menuBar().addMenu(self.viewMenu)

        self.addMenuItem(self.viewMenu, 'input', self.toggleInputVisible,
            checkcall=lambda: self.stdio.stdInputPanel.isVisible())
            
        self.addMenuItem(self.viewMenu, 'Auto Scroll', self.toggleAutoScroll,
            checkcall=lambda: self.stdio.stdOutputPanel.auto_scroll)
            
        self.addMenuItem(self.viewMenu, 'Clear', self.clear)

        scripMenu = self.menuBar().addMenu("&Script")
        self.addMenuItem(scripMenu, 'Edit sys.path...', self.editSysPaths,
            statusTip="Edit the search path used to import Python packages",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_view_list.png')))
        self.addMenuItem(scripMenu, 'Edit Live paths...', self.editLivePaths,
            statusTip="Edit the search path used to import user live scripts",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'script_gear.png')))

        self.addBaseMenu()

    def createStatusBar(self):
        self.tasktype = QLabel('')
        self.pid = QLabel('Pid:0')
        self.pname = QLabel('Name:')
        self.tid = QLabel('Tid:0')
        self.pmem = QLabel('Mem:')

        self.statusBar().addWidget(self.tasktype,1)
        self.statusBar().addWidget(self.pid,1)
        self.statusBar().addWidget(self.pname,1)
        self.statusBar().addWidget(self.tid,1)
        self.statusBar().addWidget(self.pmem,1)


    def refresh_pid_tid(self):
        self.pid.setText(f'Pid:{self.stdio.task.process_id}')
        self.tid.setText(f'Tid:{self.stdio.task.thread_id}')
        self.long_title = f'{self.short_title} Pid {self.stdio.task.process_id} Tid {self.stdio.task.thread_id}'

    def updateProcessInfo(self):
        if self.stdio.task.process_id == -1:
            return

        try:
            proc = psutil.Process(self.stdio.task.process_id)
        except:
            return

        self.tasktype.setText(f'{self.stdio.task.tasktype}')
        self.pname.setText(f'Name:{proc.name()}')
        self.tid.setText(f'Tid:{self.stdio.task.thread_id}/{proc.num_threads()}')
        self.pmem.setText(f'Mem:{proc.memory_full_info().rss / 1024 / 1024:.4g} MB')

    def set_mode(self, mode):
        self.stdio.stdInputPanel.set_mode(mode)

    def newThread(self):
        self.duplicate()

    def checkAlive(self):
        def response(mode, error_code, result):
            messageBox(f'Mode: {mode}\nError Code: {error_code}\nResult: {result}', 'Info', 'Info')

        self.stdio.task.flow_alive(response, 5)

    def suspendResumeProcess(self):
        proc = psutil.Process(self.stdio.task.process_id)

        if proc.status() == 'running':
            proc.suspend()

        elif proc.status() == 'stopped':
            proc.resume()

    def killProcess(self):
        proc = psutil.Process(self.stdio.task.process_id)

        mem = proc.memory_full_info().rss / 1024 / 1024

        if gui.dialog.question(f'Kill this following Process?\nExecutable: {proc.exe()}\nProcess id: {proc.pid} Cpu: {proc.cpu_percent(0.1)}% Mem: {mem:.4g}MB'):
            self.stdio.task.kill()

    def syncBreak(self):
        self.stdio.task.sync_break()

    def asyncBreak(self):
        self.stdio.task.async_break()

    def exec_cmd(self, cmd):
        self.stdio.stdInputPanel.execute_commands(cmd)

    def exec_file(self):
        cmd = """exec(open(gui.getfilename('*.py'),'r').read())"""
        self.exec_cmd(cmd)
        
    def openFileDialog(self):
        filepath = gui.dialog.getfilename('*.*')
        self.openFile(filepath)

    def openFile(self, filepath):
        filepath = Path(filepath)
        
        for proxyname, proxy in gui.proxies.items():
            if filepath.suffix.lower() in proxy.opens_with:
                #cmd = f"""exec(open(r'{filepath}', 'r').read())"""
                #self.exec_cmd(cmd)
                proxy.open(filepath)
                break
        
        else:
            logger.error(f'Suffix {filepath.suffix} of {filepath.name} not registered')
            return
        
        gui.qapp.history.storepath(str(filepath), category='console')        

    def addText(self, text):
        self.stdio.stdOutputPanel.addText(text)

    def editSysPaths(self):
        task = self.stdio.task
        result = task.call_func(Shell.get_sys_paths, args=(False,), wait=True)
        paths = result.copy()

        dialog_code = EditPaths(paths).exec_()

        result = task.call_func(Shell.set_sys_paths, args=(paths,))

    def editLivePaths(self):
        task = self.stdio.task

        result = task.call_func(Shell.get_live_paths, wait=True)
        paths = result.copy()

        dialog_code = EditPaths(paths).exec_()

        result = task.call_func(Shell.set_live_paths, args=(paths,))

    def toggleInputVisible(self):
        if self.stdio.stdInputPanel.isVisible():
            self.stdio.stdInputPanel.hide()
        else:
            self.stdio.stdInputPanel.show()            
        
    def toggleAutoScroll(self):
        self.stdio.stdOutputPanel.auto_scroll = not self.stdio.stdOutputPanel.auto_scroll
            
    def clear(self):
        self.stdio.stdOutputPanel.clear()

    def close_panel(self):
        try:
           self.stdio.task.system_exit()

        finally:
            self.stdio.task.unregister()
            super().close_panel()

class MainThreadConsole(Console):
    panelShortName = 'main'
    userVisible = False

    def __init__(self, mainWindow, panid):
        shell = QApplication.instance().shell
        task = tasks.ThreadTask(shell, new_thread=False)
        super().__init__(mainWindow, panid, task)
        task.start()
        self.stdio.stdInputPanel.styles['interprete'] = "background-color:#CBE9FF;"
        self.stdio.stdInputPanel.set_mode('interprete')
        self.stdio.stdInputPanel.heightHint = 0
        self.stdio.stdInputPanel.setPlainText('# Reserved for debugging only')
        self.stdio.stdInputPanel.hide()
        self.executionMenu.setEnabled(False)

        getMenuAction(self.menuBar(), ['File', 'Close']).setEnabled(False)

    def duplicate(self, floating=False):
        newpanel = gui.qapp.panels.new_panel(SubThreadConsole, None, None, floating=floating)
        return newpanel

    def close_panel(self):
        pass

class SubThreadConsole(Console):
    panelShortName = 'thread'
    userVisible = True

    def __init__(self, mainWindow, panid):
        shell = QApplication.instance().shell
        task = tasks.ThreadTask(shell, new_thread=True)
        super().__init__(mainWindow, panid, task)
        task.start()


class ChildProcessConsole(Console):
    panelShortName = 'child'
    userVisible = True

    def __init__(self, mainWindow, panid, cqs=None):
        shell = QApplication.instance().shell
        task = tasks.ProcessTask(shell, cqs)
        super().__init__(mainWindow, panid, task)
        task.start()

    def duplicate(self, floating=False):
        newpanel = gui.qapp.panels.new_panel(ChildThreadConsole, None, None, floating=floating,
            kwargs={'parent_pid': self.task.process_id})
        return newpanel


class ChildThreadConsole(Console):
    panelShortName = 'child-thread'
    userVisible = True

    def __init__(self, mainWindow, panid, parent_pid=None):
        if parent_pid is None:
            pids = [str(pid) for pid in tasks.PROCESSES.keys()][1:]
            qtypes = ['zmq', 'pipe']
            form = [('Process Id', [1] + pids), ('Queue Type', [1] + qtypes)]
            (ind0, ind1) = fedit(form)
            pid = int(pids[ind0-1])
            queue_type = qtypes[ind1-1]
        else:
            pid = parent_pid
            queue_type = 'zmq'

        master = next(iter(tasks.PROCESSES[pid].values()))
        shell = QApplication.instance().shell
        task = tasks.ProcessThreadTask(shell, master, queue_type)
        super().__init__(mainWindow, panid, task)
        task.start()

    def duplicate(self, floating=False):
        newpanel = gui.qapp.panels.new_panel(ChildThreadConsole, None, None, floating=floating,
            kwargs={'parent_pid': self.task.process_id})
        return newpanel

